import copy
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from uuid import UUID

from pydantic import (
    AfterValidator,
    AnyHttpUrl,
    ConfigDict,
    Field,
    RootModel,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated, Self

from prefect.logging import get_logger
from prefect.server.events.schemas.labelling import Labelled
from prefect.server.utilities.schemas import PrefectBaseModel
from prefect.settings import (
    PREFECT_EVENTS_MAXIMUM_LABELS_PER_RESOURCE,
    PREFECT_EVENTS_MAXIMUM_RELATED_RESOURCES,
)
from prefect.types import DateTime
from prefect.types._datetime import now

if TYPE_CHECKING:
    import logging

    logger: "logging.Logger" = get_logger(__name__)


class Resource(Labelled):
    """An observable business object of interest to the user"""

    @model_validator(mode="after")
    def enforce_maximum_labels(self) -> Self:
        if len(self.root) > PREFECT_EVENTS_MAXIMUM_LABELS_PER_RESOURCE.value():
            raise ValueError(
                "The maximum number of labels per resource "
                f"is {PREFECT_EVENTS_MAXIMUM_LABELS_PER_RESOURCE.value()}"
            )

        return self

    @model_validator(mode="after")
    def requires_resource_id(self) -> Self:
        if "prefect.resource.id" not in self.root:
            raise ValueError("Resources must include the prefect.resource.id label")
        if not self.root["prefect.resource.id"]:
            raise ValueError("The prefect.resource.id label must be non-empty")

        return self

    @property
    def id(self) -> str:
        return self["prefect.resource.id"]

    @property
    def name(self) -> Optional[str]:
        return self.get("prefect.resource.name")

    def prefect_object_id(self, kind: str) -> UUID:
        """Extracts the UUID from an event's resource ID if it's the expected kind
        of prefect resource"""
        prefix = f"{kind}." if not kind.endswith(".") else kind

        if not self.id.startswith(prefix):
            raise ValueError(f"Resource ID {self.id} does not start with {prefix}")

        return UUID(self.id[len(prefix) :])


class RelatedResource(Resource):
    """A Resource with a specific role in an Event"""

    @model_validator(mode="after")
    def requires_resource_role(self) -> Self:
        if "prefect.resource.role" not in self.root:
            raise ValueError(
                "Related Resources must include the prefect.resource.role label"
            )
        if not self.root["prefect.resource.role"]:
            raise ValueError("The prefect.resource.role label must be non-empty")

        return self

    @property
    def role(self) -> str:
        return self["prefect.resource.role"]


def _validate_event_name_length(value: str) -> str:
    from prefect.settings import PREFECT_SERVER_EVENTS_MAXIMUM_EVENT_NAME_LENGTH

    if len(value) > PREFECT_SERVER_EVENTS_MAXIMUM_EVENT_NAME_LENGTH.value():
        raise ValueError(
            f"Event name must be at most {PREFECT_SERVER_EVENTS_MAXIMUM_EVENT_NAME_LENGTH.value()} characters"
        )
    return value


class Event(PrefectBaseModel):
    """The client-side view of an event that has happened to a Resource"""

    occurred: DateTime = Field(
        description="When the event happened from the sender's perspective",
    )
    event: Annotated[str, AfterValidator(_validate_event_name_length)] = Field(
        description="The name of the event that happened",
    )
    resource: Resource = Field(
        description="The primary Resource this event concerns",
    )
    related: list[RelatedResource] = Field(
        default_factory=list,
        description="A list of additional Resources involved in this event",
    )
    payload: dict[str, Any] = Field(
        default_factory=dict,
        description="An open-ended set of data describing what happened",
    )
    id: UUID = Field(
        description="The client-provided identifier of this event",
    )
    follows: Optional[UUID] = Field(
        default=None,
        description=(
            "The ID of an event that is known to have occurred prior to this one. "
            "If set, this may be used to establish a more precise ordering of causally-"
            "related events when they occur close enough together in time that the "
            "system may receive them out-of-order."
        ),
    )

    @property
    def involved_resources(self) -> Sequence[Resource]:
        return [self.resource] + list(self.related)

    @property
    def resource_in_role(self) -> Mapping[str, RelatedResource]:
        """Returns a mapping of roles to the first related resource in that role"""
        return {related.role: related for related in reversed(self.related)}

    @property
    def resources_in_role(self) -> Mapping[str, Sequence[RelatedResource]]:
        """Returns a mapping of roles to related resources in that role"""
        resources: Dict[str, List[RelatedResource]] = defaultdict(list)
        for related in self.related:
            resources[related.role].append(related)
        return resources

    @field_validator("related")
    @classmethod
    def enforce_maximum_related_resources(
        cls, value: List[RelatedResource]
    ) -> List[RelatedResource]:
        if len(value) > PREFECT_EVENTS_MAXIMUM_RELATED_RESOURCES.value():
            raise ValueError(
                "The maximum number of related resources "
                f"is {PREFECT_EVENTS_MAXIMUM_RELATED_RESOURCES.value()}"
            )

        return value

    def receive(self, received: Optional[DateTime] = None) -> "ReceivedEvent":
        kwargs = self.model_dump()
        if received is not None:
            kwargs["received"] = received
        return ReceivedEvent(**kwargs)

    def find_resource_label(self, label: str) -> Optional[str]:
        """Finds the value of the given label in this event's resource or one of its
        related resources.  If the label starts with `related:<role>:`, search for the
        first matching label in a related resource with that role."""
        directive, _, related_label = label.rpartition(":")
        directive, _, role = directive.partition(":")
        if directive == "related":
            for related in self.related:
                if related.role == role:
                    return related.get(related_label)
        return self.resource.get(label)


class ReceivedEvent(Event):
    """The server-side view of an event that has happened to a Resource after it has
    been received by the server"""

    model_config: ClassVar[ConfigDict] = ConfigDict(
        extra="ignore", from_attributes=True
    )

    received: DateTime = Field(
        default_factory=lambda: now("UTC"),
        description="When the event was received by Prefect Cloud",
    )

    def as_database_row(self) -> dict[str, Any]:
        row = self.model_dump()
        row["resource_id"] = self.resource.id
        row["recorded"] = now("UTC")
        row["related_resource_ids"] = [related.id for related in self.related]
        return row

    def as_database_resource_rows(self) -> List[Dict[str, Any]]:
        def without_id_and_role(resource: Resource) -> Dict[str, str]:
            d: Dict[str, str] = resource.root.copy()
            d.pop("prefect.resource.id", None)
            d.pop("prefect.resource.role", None)
            return d

        return [
            {
                "occurred": self.occurred,
                "resource_id": resource.id,
                "resource_role": (
                    resource.role if isinstance(resource, RelatedResource) else ""
                ),
                "resource": without_id_and_role(resource),
                "event_id": self.id,
            }
            for resource in [self.resource, *self.related]
        ]


def matches(expected: str, value: Optional[str]) -> bool:
    """Returns true if the given value matches the expected string, which may
    include a a negation prefix ("!this-value") or a wildcard suffix
    ("any-value-starting-with*")"""
    if value is None:
        return False

    positive = True
    if expected.startswith("!"):
        expected = expected[1:]
        positive = False

    if expected.endswith("*"):
        match = value.startswith(expected[:-1])
    else:
        match = value == expected

    return match if positive else not match


class ResourceSpecification(RootModel[Dict[str, Union[str, List[str]]]]):
    def matches_every_resource(self) -> bool:
        return len(self.root) == 0

    def matches_every_resource_of_kind(self, prefix: str) -> bool:
        if self.matches_every_resource():
            return True
        if len(self.root) == 1:
            resource_id = self.root.get("prefect.resource.id")
            if resource_id:
                values = [resource_id] if isinstance(resource_id, str) else resource_id
                return any(value == f"{prefix}.*" for value in values)
        return False

    def includes(self, candidates: Iterable[Resource]) -> bool:
        if self.matches_every_resource():
            return True
        for candidate in candidates:
            if self.matches(candidate):
                return True
        return False

    def matches(self, resource: Resource) -> bool:
        for label, expected in self.items():
            value = resource.get(label)
            if not any(matches(candidate, value) for candidate in expected):
                return False
        return True

    def items(self) -> Iterable[Tuple[str, List[str]]]:
        return [
            (label, [value] if isinstance(value, str) else value)
            for label, value in self.root.items()
        ]

    def __contains__(self, key: str) -> bool:
        return key in self.root

    def __getitem__(self, key: str) -> List[str]:
        value = self.root[key]
        if not value:
            return []
        if not isinstance(value, list):
            value = [value]
        return value

    def pop(
        self, key: str, default: Optional[Union[str, List[str]]] = None
    ) -> Optional[List[str]]:
        value = self.root.pop(key, default)
        if not value:
            return []
        if not isinstance(value, list):
            value = [value]
        return value

    def get(
        self, key: str, default: Optional[Union[str, List[str]]] = None
    ) -> Optional[List[str]]:
        value = self.root.get(key, default)
        if not value:
            return []
        if not isinstance(value, list):
            value = [value]
        return value

    def __len__(self) -> int:
        return len(self.root)

    def deepcopy(self) -> "ResourceSpecification":
        return ResourceSpecification(root=copy.deepcopy(self.root))


class EventPage(PrefectBaseModel):
    """A single page of events returned from the API, with an optional link to the
    next page of results"""

    events: List[ReceivedEvent] = Field(
        ..., description="The Events matching the query"
    )
    total: int = Field(..., description="The total number of matching Events")
    next_page: Optional[AnyHttpUrl] = Field(
        ..., description="The URL for the next page of results, if there are more"
    )


class EventCount(PrefectBaseModel):
    """The count of events with the given filter value"""

    value: str = Field(..., description="The value to use for filtering")
    label: str = Field(..., description="The value to display for this count")
    count: int = Field(..., description="The count of matching events")
    start_time: DateTime = Field(
        ..., description="The start time of this group of events"
    )
    end_time: DateTime = Field(..., description="The end time of this group of events")



"""
This module contains functionality for creating schedules for deployments.
"""

from __future__ import annotations

import dataclasses
import datetime
from functools import partial
from typing import Any

from prefect._internal.schemas.validators import (
    validate_cron_string,
    validate_rrule_string,
)


@dataclasses.dataclass(frozen=True)
class Schedule:
    """
    A dataclass representing a schedule.

    Note that only one of `interval`, `cron`, or `rrule` can be defined at a time.

    Attributes:
        interval: A timedelta representing the frequency of the schedule.
        cron: A valid cron string (e.g. "0 0 * * *").
        rrule: A valid RRule string (e.g. "RRULE:FREQ=DAILY;INTERVAL=1").
        timezone: A valid timezone string in IANA tzdata format (e.g. America/New_York).
        anchor_date: An anchor date to schedule increments against; if not provided,
            the current timestamp will be used.
        day_or: Control how `day` and `day_of_week` entries are handled.
            Defaults to True, matching cron which connects those values using
            OR. If the switch is set to False, the values are connected using AND.
            This behaves like fcron and enables you to e.g. define a job that
            executes each 2nd friday of a month by setting the days of month and
            the weekday.
        active: Whether or not the schedule is active.
        parameters: A dictionary containing parameter overrides for the schedule.
        slug: A unique identifier for the schedule.
    """

    interval: datetime.timedelta | None = None
    cron: str | None = None
    rrule: str | None = None
    timezone: str | None = None
    anchor_date: datetime.datetime = dataclasses.field(
        default_factory=partial(datetime.datetime.now, tz=datetime.timezone.utc)
    )
    day_or: bool = False
    active: bool = True
    parameters: dict[str, Any] = dataclasses.field(default_factory=dict)
    slug: str | None = None

    def __post_init__(self) -> None:
        defined_fields = [
            field
            for field in ["interval", "cron", "rrule"]
            if getattr(self, field) is not None
        ]
        if len(defined_fields) > 1:
            raise ValueError(
                f"Only one schedule type can be defined at a time. Found: {', '.join(defined_fields)}"
            )

        if self.cron is not None:
            validate_cron_string(self.cron)
        if self.rrule is not None:
            validate_rrule_string(self.rrule)


def Cron(
    cron: str,
    /,
    timezone: str | None = None,
    day_or: bool = False,
    active: bool = True,
    parameters: dict[str, Any] | None = None,
    slug: str | None = None,
) -> Schedule:
    """
    Creates a cron schedule.

    Args:
        cron: A valid cron string (e.g. "0 0 * * *").
        timezone: A valid timezone string in IANA tzdata format (e.g. America/New_York).
        day_or: Control how `day` and `day_of_week` entries are handled.
            Defaults to True, matching cron which connects those values using
            OR. If the switch is set to False, the values are connected using AND.
            This behaves like fcron and enables you to e.g. define a job that
            executes each 2nd friday of a month by setting the days of month and
            the weekday.
        active: Whether or not the schedule is active.
        parameters: A dictionary containing parameter overrides for the schedule.
        slug: A unique identifier for the schedule.

    Returns:
        A cron schedule.

    Examples:
        Create a cron schedule that runs every day at 12:00 AM UTC:
        ```python
        from prefect.schedules import Cron

        Cron("0 0 * * *")
        ```

        Create a cron schedule that runs every Monday at 8:00 AM in the America/New_York timezone:
        ```python
        from prefect.schedules import Cron

        Cron("0 8 * * 1", timezone="America/New_York")
        ```

    """
    if parameters is None:
        parameters = {}
    return Schedule(
        cron=cron,
        timezone=timezone,
        day_or=day_or,
        active=active,
        parameters=parameters,
        slug=slug,
    )


def Interval(
    interval: datetime.timedelta | float | int,
    /,
    anchor_date: datetime.datetime | None = None,
    timezone: str | None = None,
    active: bool = True,
    parameters: dict[str, Any] | None = None,
    slug: str | None = None,
) -> Schedule:
    """
    Creates an interval schedule.

    Args:
        interval: The interval to use for the schedule. If an integer is provided,
            it will be interpreted as seconds.
        anchor_date: The anchor date to use for the schedule.
        timezone: A valid timezone string in IANA tzdata format (e.g. America/New_York).
        active: Whether or not the schedule is active.
        parameters: A dictionary containing parameter overrides for the schedule.
        slug: A unique identifier for the schedule.

    Returns:
        An interval schedule.

    Examples:
        Create an interval schedule that runs every hour:
        ```python
        from datetime import timedelta

        from prefect.schedules import Interval

        Interval(timedelta(hours=1))
        ```

        Create an interval schedule that runs every 60 seconds starting at a specific date:
        ```python
        from datetime import datetime

        from prefect.schedules import Interval

        Interval(60, anchor_date=datetime(2024, 1, 1))
        ```
    """
    if isinstance(interval, (float, int)):
        interval = datetime.timedelta(seconds=interval)
    if anchor_date is None:
        anchor_date = datetime.datetime.now(tz=datetime.timezone.utc)
    if parameters is None:
        parameters = {}
    return Schedule(
        interval=interval,
        anchor_date=anchor_date,
        timezone=timezone,
        active=active,
        parameters=parameters,
        slug=slug,
    )


def RRule(
    rrule: str,
    /,
    timezone: str | None = None,
    active: bool = True,
    parameters: dict[str, Any] | None = None,
    slug: str | None = None,
) -> Schedule:
    """
    Creates an RRule schedule.

    Args:
        rrule: A valid RRule string (e.g. "RRULE:FREQ=DAILY;INTERVAL=1").
        timezone: A valid timezone string in IANA tzdata format (e.g. America/New_York).
        active: Whether or not the schedule is active.
        parameters: A dictionary containing parameter overrides for the schedule.
        slug: A unique identifier for the schedule.

    Returns:
        An RRule schedule.

    Examples:
        Create an RRule schedule that runs every day at 12:00 AM UTC:
        ```python
        from prefect.schedules import RRule

        RRule("RRULE:FREQ=DAILY;INTERVAL=1")
        ```

        Create an RRule schedule that runs every 2nd friday of the month in the America/Chicago timezone:
        ```python
        from prefect.schedules import RRule

        RRule("RRULE:FREQ=MONTHLY;INTERVAL=1;BYDAY=2FR", timezone="America/Chicago")
        ```
    """
    if parameters is None:
        parameters = {}
    return Schedule(
        rrule=rrule,
        timezone=timezone,
        active=active,
        parameters=parameters,
        slug=slug,
    )



from __future__ import annotations

import abc
from datetime import timedelta
from typing import Literal, Optional, Union
from uuid import UUID

from pydantic import Field, PrivateAttr, computed_field
from typing_extensions import Self, TypeAlias

from prefect._internal.schemas.bases import PrefectBaseModel


class ServiceLevelAgreement(PrefectBaseModel, abc.ABC):
    """An ORM representation of a Service Level Agreement."""

    _deployment_id: Optional[UUID] = PrivateAttr(default=None)

    name: str = Field(
        default=...,
        description="The name of the SLA. Names must be unique on a per-deployment basis.",
    )
    severity: Literal["minor", "low", "moderate", "high", "critical"] = Field(
        default="moderate",
        description="The severity of the SLA.",
    )
    enabled: Optional[bool] = Field(
        default=True,
        description="Whether the SLA is enabled.",
    )

    def set_deployment_id(self, deployment_id: UUID) -> Self:
        self._deployment_id = deployment_id
        return self

    @computed_field
    @property
    def owner_resource(self) -> Union[str, None]:
        if self._deployment_id:
            return f"prefect.deployment.{self._deployment_id}"
        return None


class TimeToCompletionSla(ServiceLevelAgreement):
    """An SLA that triggers when a flow run takes longer than the specified duration."""

    duration: int = Field(
        default=...,
        description="The maximum flow run duration allowed before the SLA is violated, expressed in seconds.",
    )


class FrequencySla(ServiceLevelAgreement):
    """An SLA that triggers when a completed flow run is not detected in the specified time.

    For example, if stale_after is 1 hour, if a flow run does not complete
    within an hour of the previous flow run, the SLA will trigger.
    """

    stale_after: timedelta = Field(
        default=...,
        description="The amount of time after which a flow run is considered in violation.",
    )


class LatenessSla(ServiceLevelAgreement):
    """An SLA that triggers when a flow run does not start within the specified window.

    For example, if you schedule the deployment to run every day at 2:00pm and you pass
    within=timedelta(minutes=10) to this SLA, if a run hasn't started by 2:10pm the SLA
    violation will be recorded.
    """

    within: timedelta = Field(
        default=...,
        description="The amount of time before a flow run is considered in violation.",
    )


class SlaMergeResponse(PrefectBaseModel):
    """A response object for the apply_slas_for_deployment method. Contains the names of the created, updated, and deleted SLAs."""

    created: list[str]
    updated: list[str]
    deleted: list[str]


# Concrete SLA types
SlaTypes: TypeAlias = Union[TimeToCompletionSla, LatenessSla, FrequencySla]



from collections.abc import Generator
from contextlib import contextmanager
from typing import Optional, TypeVar, Union

from typing_extensions import Literal

from prefect.client.schemas.responses import MinimalConcurrencyLimitResponse
from prefect.types._datetime import now
from prefect.utilities.asyncutils import run_coro_as_sync

from ._asyncio import (
    aacquire_concurrency_slots,
    arelease_concurrency_slots,
)
from ._events import (
    emit_concurrency_acquisition_events,
    emit_concurrency_release_events,
)

T = TypeVar("T")


def _release_concurrency_slots(
    names: list[str], slots: int, occupancy_seconds: float
) -> list[MinimalConcurrencyLimitResponse]:
    result = run_coro_as_sync(
        arelease_concurrency_slots(names, slots, occupancy_seconds)
    )
    return result


def _acquire_concurrency_slots(
    names: list[str],
    slots: int,
    mode: Literal["concurrency", "rate_limit"] = "concurrency",
    timeout_seconds: Optional[float] = None,
    create_if_missing: Optional[bool] = None,
    max_retries: Optional[int] = None,
    strict: bool = False,
) -> list[MinimalConcurrencyLimitResponse]:
    result = run_coro_as_sync(
        aacquire_concurrency_slots(
            names, slots, mode, timeout_seconds, create_if_missing, max_retries, strict
        )
    )
    return result


@contextmanager
def concurrency(
    names: Union[str, list[str]],
    occupy: int = 1,
    timeout_seconds: Optional[float] = None,
    max_retries: Optional[int] = None,
    strict: bool = False,
    create_if_missing: Optional[bool] = None,
) -> Generator[None, None, None]:
    """A context manager that acquires and releases concurrency slots from the
    given concurrency limits.

    Args:
        names: The names of the concurrency limits to acquire slots from.
        occupy: The number of slots to acquire and hold from each limit.
        timeout_seconds: The number of seconds to wait for the slots to be acquired before
            raising a `TimeoutError`. A timeout of `None` will wait indefinitely.
        max_retries: The maximum number of retries to acquire the concurrency slots.
        strict: A boolean specifying whether to raise an error if the concurrency limit does not exist.
            Defaults to `False`.

    Raises:
        TimeoutError: If the slots are not acquired within the given timeout.
        ConcurrencySlotAcquisitionError: If the concurrency limit does not exist and `strict` is `True`.

    Example:
    A simple example of using the sync `concurrency` context manager:
    ```python
    from prefect.concurrency.sync import concurrency

    def resource_heavy():
        with concurrency("test", occupy=1):
            print("Resource heavy task")

    def main():
        resource_heavy()
    ```
    """
    if not names:
        yield
        return

    names = names if isinstance(names, list) else [names]

    limits: list[MinimalConcurrencyLimitResponse] = _acquire_concurrency_slots(
        names,
        occupy,
        timeout_seconds=timeout_seconds,
        create_if_missing=create_if_missing,
        strict=strict,
        max_retries=max_retries,
    )
    acquisition_time = now("UTC")
    emitted_events = emit_concurrency_acquisition_events(limits, occupy)

    try:
        yield
    finally:
        occupancy_period = now("UTC") - acquisition_time
        _release_concurrency_slots(names, occupy, occupancy_period.total_seconds())
        emit_concurrency_release_events(limits, occupy, emitted_events)


def rate_limit(
    names: Union[str, list[str]],
    occupy: int = 1,
    timeout_seconds: Optional[float] = None,
    create_if_missing: Optional[bool] = None,
    strict: bool = False,
) -> None:
    """Block execution until an `occupy` number of slots of the concurrency
    limits given in `names` are acquired. Requires that all given concurrency
    limits have a slot decay.

    Args:
        names: The names of the concurrency limits to acquire slots from.
        occupy: The number of slots to acquire and hold from each limit.
        timeout_seconds: The number of seconds to wait for the slots to be acquired before
            raising a `TimeoutError`. A timeout of `None` will wait indefinitely.
        strict: A boolean specifying whether to raise an error if the concurrency limit does not exist.
            Defaults to `False`.

    Raises:
        TimeoutError: If the slots are not acquired within the given timeout.
        ConcurrencySlotAcquisitionError: If the concurrency limit does not exist and `strict` is `True`.
    """
    if not names:
        return

    names = names if isinstance(names, list) else [names]

    limits = _acquire_concurrency_slots(
        names,
        occupy,
        mode="rate_limit",
        timeout_seconds=timeout_seconds,
        create_if_missing=create_if_missing,
        strict=strict,
    )
    emit_concurrency_acquisition_events(limits, occupy)



from pathlib import Path
from typing import Any, Optional

from prefect.settings import (
    PREFECT_DEFAULT_DOCKER_BUILD_NAMESPACE,
)
from prefect.types._datetime import now
from prefect.utilities.dockerutils import (
    PushError,
    build_image,
    docker_client,
    generate_default_dockerfile,
    parse_image_tag,
    split_repository_path,
)
from prefect.utilities.slugify import slugify


class DockerImage:
    """
    Configuration used to build and push a Docker image for a deployment.

    Attributes:
        name: The name of the Docker image to build, including the registry and
            repository.
        tag: The tag to apply to the built image.
        dockerfile: The path to the Dockerfile to use for building the image. If
            not provided, a default Dockerfile will be generated.
        **build_kwargs: Additional keyword arguments to pass to the Docker build request.
            See the [`docker-py` documentation](https://docker-py.readthedocs.io/en/stable/images.html#docker.models.images.ImageCollection.build)
            for more information.

    """

    def __init__(
        self,
        name: str,
        tag: Optional[str] = None,
        dockerfile: str = "auto",
        **build_kwargs: Any,
    ):
        image_name, image_tag = parse_image_tag(name)
        if tag and image_tag:
            raise ValueError(
                f"Only one tag can be provided - both {image_tag!r} and {tag!r} were"
                " provided as tags."
            )
        namespace, repository = split_repository_path(image_name)
        # if the provided image name does not include a namespace (registry URL or user/org name),
        # use the default namespace
        if not namespace:
            namespace = PREFECT_DEFAULT_DOCKER_BUILD_NAMESPACE.value()
        # join the namespace and repository to create the full image name
        # ignore namespace if it is None
        self.name: str = "/".join(filter(None, [namespace, repository]))
        self.tag: str = tag or image_tag or slugify(now("UTC").isoformat())
        self.dockerfile: str = dockerfile
        self.build_kwargs: dict[str, Any] = build_kwargs

    @property
    def reference(self) -> str:
        return f"{self.name}:{self.tag}"

    def build(self) -> None:
        full_image_name = self.reference
        build_kwargs = self.build_kwargs.copy()
        if "context" not in build_kwargs:
            build_kwargs["context"] = Path.cwd()
        build_kwargs["tag"] = full_image_name
        build_kwargs["pull"] = build_kwargs.get("pull", True)

        if self.dockerfile == "auto":
            with generate_default_dockerfile():
                build_image(**build_kwargs)
        else:
            build_kwargs["dockerfile"] = self.dockerfile
            build_image(**build_kwargs)

    def push(self) -> None:
        with docker_client() as client:
            events = client.api.push(
                repository=self.name, tag=self.tag, stream=True, decode=True
            )
            for event in events:
                if "error" in event:
                    raise PushError(event["error"])



"""Add parameters column to deployment schedules

Revision ID: 67f886da208e
Revises: a49711513ad4
Create Date: 2025-02-03 12:52:28.370958

"""

import sqlalchemy as sa
from alembic import op

import prefect

# revision identifiers, used by Alembic.
revision = "67f886da208e"
down_revision = "a49711513ad4"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "deployment_schedule",
        sa.Column(
            "parameters",
            prefect.server.utilities.database.JSON,
            server_default="{}",
            nullable=False,
        ),
    )


def downgrade():
    op.drop_column("deployment_schedule", "parameters")



import math
import random


def poisson_interval(
    average_interval: float, lower: float = 0, upper: float = 1
) -> float:
    """
    Generates an "inter-arrival time" for a Poisson process.

    Draws a random variable from an exponential distribution using the inverse-CDF
    method. Can optionally be passed a lower and upper bound between (0, 1] to clamp
    the potential output values.
    """

    # note that we ensure the argument to the logarithm is stabilized to prevent
    # calling log(0), which results in a DomainError
    return -math.log(max(1 - random.uniform(lower, upper), 1e-10)) * average_interval


def exponential_cdf(x: float, average_interval: float) -> float:
    ld = 1 / average_interval
    return 1 - math.exp(-ld * x)


def lower_clamp_multiple(k: float) -> float:
    """
    Computes a lower clamp multiple that can be used to bound a random variate drawn
    from an exponential distribution.

    Given an upper clamp multiple `k` (and corresponding upper bound k * average_interval),
    this function computes a lower clamp multiple `c` (corresponding to a lower bound
    c * average_interval) where the probability mass between the lower bound and the
    median is equal to the probability mass between the median and the upper bound.
    """
    if k >= 50:
        # return 0 for large values of `k` to prevent numerical overflow
        return 0.0

    return math.log(max(2**k / (2**k - 1), 1e-10), 2)


def clamped_poisson_interval(
    average_interval: float, clamping_factor: float = 0.3
) -> float:
    """
    Bounds Poisson "inter-arrival times" to a range defined by the clamping factor.

    The upper bound for this random variate is: average_interval * (1 + clamping_factor).
    A lower bound is picked so that the average interval remains approximately fixed.
    """
    if clamping_factor <= 0:
        raise ValueError("`clamping_factor` must be >= 0.")

    upper_clamp_multiple = 1 + clamping_factor
    upper_bound = average_interval * upper_clamp_multiple
    lower_bound = max(0, average_interval * lower_clamp_multiple(upper_clamp_multiple))

    upper_rv = exponential_cdf(upper_bound, average_interval)
    lower_rv = exponential_cdf(lower_bound, average_interval)
    return poisson_interval(average_interval, lower_rv, upper_rv)


def bounded_poisson_interval(lower_bound: float, upper_bound: float) -> float:
    """
    Bounds Poisson "inter-arrival times" to a range.

    Unlike `clamped_poisson_interval` this does not take a target average interval.
    Instead, the interval is predetermined and the average is calculated as their
    midpoint. This allows Poisson intervals to be used in cases where a lower bound
    must be enforced.
    """
    average = (float(lower_bound) + float(upper_bound)) / 2.0
    upper_rv = exponential_cdf(upper_bound, average)
    lower_rv = exponential_cdf(lower_bound, average)
    return poisson_interval(average, lower_rv, upper_rv)



"""
The MarkLateRuns service. Responsible for putting flow runs in a Late state if they are not started on time.
The threshold for a late run can be configured by changing `PREFECT_API_SERVICES_LATE_RUNS_AFTER_SECONDS`.
"""

from __future__ import annotations

import asyncio
import datetime
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession

import prefect.server.models as models
from prefect.server.database import PrefectDBInterface, inject_db
from prefect.server.database.dependencies import db_injector
from prefect.server.exceptions import ObjectNotFoundError
from prefect.server.orchestration.core_policy import MarkLateRunsPolicy
from prefect.server.schemas import states
from prefect.server.services.base import LoopService
from prefect.settings import (
    PREFECT_API_SERVICES_LATE_RUNS_AFTER_SECONDS,
    PREFECT_API_SERVICES_LATE_RUNS_LOOP_SECONDS,
)
from prefect.settings.context import get_current_settings
from prefect.settings.models.server.services import ServicesBaseSetting
from prefect.types._datetime import DateTime, now

if TYPE_CHECKING:
    from uuid import UUID


class MarkLateRuns(LoopService):
    """
    Finds flow runs that are later than their scheduled start time

    A flow run is defined as "late" if has not scheduled within a certain amount
    of time after its scheduled start time. The exact amount is configurable in
    Prefect REST API Settings.
    """

    @classmethod
    def service_settings(cls) -> ServicesBaseSetting:
        return get_current_settings().server.services.late_runs

    def __init__(self, loop_seconds: float | None = None, **kwargs: Any):
        super().__init__(
            loop_seconds=loop_seconds
            or PREFECT_API_SERVICES_LATE_RUNS_LOOP_SECONDS.value(),
            **kwargs,
        )

        # mark runs late if they are this far past their expected start time
        self.mark_late_after: datetime.timedelta = (
            PREFECT_API_SERVICES_LATE_RUNS_AFTER_SECONDS.value()
        )

        # query for this many runs to mark as late at once
        self.batch_size = 400

    @db_injector
    async def run_once(self, db: PrefectDBInterface) -> None:
        """
        Mark flow runs as late by:

        - Querying for flow runs in a scheduled state that are Scheduled to start in the past
        - For any runs past the "late" threshold, setting the flow run state to a new `Late` state
        """
        scheduled_to_start_before = now("UTC") - datetime.timedelta(
            seconds=self.mark_late_after.total_seconds()
        )

        while True:
            async with db.session_context(begin_transaction=True) as session:
                query = self._get_select_late_flow_runs_query(
                    scheduled_to_start_before=scheduled_to_start_before, db=db
                )

                result = await session.execute(query)
                runs = result.all()

                # mark each run as late
                for run in runs:
                    await self._mark_flow_run_as_late(session=session, flow_run=run)

                # if no runs were found, exit the loop
                if len(runs) < self.batch_size:
                    break

        self.logger.info("Finished monitoring for late runs.")

    @inject_db
    def _get_select_late_flow_runs_query(
        self, scheduled_to_start_before: datetime.datetime, db: PrefectDBInterface
    ) -> sa.Select[tuple["UUID", DateTime | None]]:
        """
        Returns a sqlalchemy query for late flow runs.

        Args:
            scheduled_to_start_before: the maximum next scheduled start time of
                scheduled flow runs to consider in the returned query
        """
        query = (
            sa.select(
                db.FlowRun.id,
                db.FlowRun.next_scheduled_start_time,
            )
            .where(
                # The next scheduled start time is in the past, including the mark late
                # after buffer
                (db.FlowRun.next_scheduled_start_time <= scheduled_to_start_before),
                db.FlowRun.state_type == states.StateType.SCHEDULED,
                db.FlowRun.state_name == "Scheduled",
            )
            .limit(self.batch_size)
        )
        return query

    async def _mark_flow_run_as_late(
        self,
        session: AsyncSession,
        flow_run: sa.Row[tuple["UUID", DateTime | None]],
    ) -> None:
        """
        Mark a flow run as late.

        Pass-through method for overrides.
        """
        try:
            await models.flow_runs.set_flow_run_state(
                session=session,
                flow_run_id=flow_run.id,
                state=states.Late(scheduled_time=flow_run.next_scheduled_start_time),
                flow_policy=MarkLateRunsPolicy,  # type: ignore
            )
        except ObjectNotFoundError:
            return  # flow run was deleted, ignore it


if __name__ == "__main__":
    asyncio.run(MarkLateRuns(handle_signals=True).start())



import hashlib
from functools import partial
from pathlib import Path
from typing import Any, Callable, Optional, Union

import cloudpickle  # type: ignore  # no stubs available

from prefect.exceptions import HashError
from prefect.serializers import JSONSerializer

_md5 = partial(hashlib.md5, usedforsecurity=False)


def stable_hash(*args: Union[str, bytes], hash_algo: Callable[..., Any] = _md5) -> str:
    """Given some arguments, produces a stable 64-bit hash of their contents.

    Supports bytes and strings. Strings will be UTF-8 encoded.

    Args:
        *args: Items to include in the hash.
        hash_algo: Hash algorithm from hashlib to use.

    Returns:
        A hex hash.
    """
    h = hash_algo()
    for a in args:
        if isinstance(a, str):
            a = a.encode()
        h.update(a)
    return h.hexdigest()


def file_hash(path: str, hash_algo: Callable[..., Any] = _md5) -> str:
    """Given a path to a file, produces a stable hash of the file contents.

    Args:
        path (str): the path to a file
        hash_algo: Hash algorithm from hashlib to use.

    Returns:
        str: a hash of the file contents
    """
    contents = Path(path).read_bytes()
    return stable_hash(contents, hash_algo=hash_algo)


def hash_objects(
    *args: Any,
    hash_algo: Callable[..., Any] = _md5,
    raise_on_failure: bool = False,
    **kwargs: Any,
) -> Optional[str]:
    """
    Attempt to hash objects by dumping to JSON or serializing with cloudpickle.

    Args:
        *args: Positional arguments to hash
        hash_algo: Hash algorithm to use
        raise_on_failure: If True, raise exceptions instead of returning None
        **kwargs: Keyword arguments to hash

    Returns:
        A hash string or None if hashing failed

    Raises:
        HashError: If objects cannot be hashed and raise_on_failure is True
    """
    json_error = None
    pickle_error = None

    try:
        serializer = JSONSerializer(dumps_kwargs={"sort_keys": True})
        return stable_hash(serializer.dumps((args, kwargs)), hash_algo=hash_algo)
    except Exception as e:
        json_error = str(e)

    try:
        return stable_hash(cloudpickle.dumps((args, kwargs)), hash_algo=hash_algo)  # type: ignore[reportUnknownMemberType]
    except Exception as e:
        pickle_error = str(e)

    if raise_on_failure:
        msg = (
            "Unable to create hash - objects could not be serialized.\n"
            f"  JSON error: {json_error}\n"
            f"  Pickle error: {pickle_error}"
        )
        raise HashError(msg)

    return None



from __future__ import annotations

import asyncio
import atexit
import threading
import uuid
from typing import (
    TYPE_CHECKING,
    Callable,
)

import anyio
from cachetools import TTLCache
from typing_extensions import Self

from prefect._internal.concurrency.api import create_call, from_async, from_sync
from prefect._internal.concurrency.threads import get_global_loop
from prefect.client.schemas.objects import (
    TERMINAL_STATES,
)
from prefect.events.clients import get_events_subscriber
from prefect.events.filters import EventFilter, EventNameFilter
from prefect.logging import get_logger

if TYPE_CHECKING:
    import logging


class FlowRunWaiter:
    """
    A service used for waiting for a flow run to finish.

    This service listens for flow run events and provides a way to wait for a specific
    flow run to finish. This is useful for waiting for a flow run to finish before
    continuing execution.

    The service is a singleton and must be started before use. The service will
    automatically start when the first instance is created. A single websocket
    connection is used to listen for flow run events.

    The service can be used to wait for a flow run to finish by calling
    `FlowRunWaiter.wait_for_flow_run` with the flow run ID to wait for. The method
    will return when the flow run has finished or the timeout has elapsed.

    The service will automatically stop when the Python process exits or when the
    global loop thread is stopped.

    Example:
    ```python
    import asyncio
    from uuid import uuid4

    from prefect import flow
    from prefect.flow_engine import run_flow_async
    from prefect.flow_runs import FlowRunWaiter


    @flow
    async def test_flow():
        await asyncio.sleep(5)
        print("Done!")


    async def main():
        flow_run_id = uuid4()
        asyncio.create_flow(run_flow_async(flow=test_flow, flow_run_id=flow_run_id))

        await FlowRunWaiter.wait_for_flow_run(flow_run_id)
        print("Flow run finished")


    if __name__ == "__main__":
        asyncio.run(main())
    ```
    """

    _instance: Self | None = None
    _instance_lock = threading.Lock()

    def __init__(self):
        self.logger: "logging.Logger" = get_logger("FlowRunWaiter")
        self._consumer_task: asyncio.Task[None] | None = None
        self._observed_completed_flow_runs: TTLCache[uuid.UUID, bool] = TTLCache(
            maxsize=10000, ttl=600
        )
        self._completion_events: dict[uuid.UUID, asyncio.Event] = {}
        self._completion_callbacks: dict[uuid.UUID, Callable[[], None]] = {}
        self._loop: asyncio.AbstractEventLoop | None = None
        self._observed_completed_flow_runs_lock = threading.Lock()
        self._completion_events_lock = threading.Lock()
        self._started = False

    def start(self) -> None:
        """
        Start the FlowRunWaiter service.
        """
        if self._started:
            return
        self.logger.debug("Starting FlowRunWaiter")
        loop_thread = get_global_loop()

        if not asyncio.get_running_loop() == loop_thread.loop:
            raise RuntimeError("FlowRunWaiter must run on the global loop thread.")

        self._loop = loop_thread.loop
        if TYPE_CHECKING:
            assert self._loop is not None

        consumer_started = asyncio.Event()
        self._consumer_task = self._loop.create_task(
            self._consume_events(consumer_started)
        )
        asyncio.run_coroutine_threadsafe(consumer_started.wait(), self._loop)

        loop_thread.add_shutdown_call(create_call(self.stop))
        atexit.register(self.stop)
        self._started = True

    async def _consume_events(self, consumer_started: asyncio.Event):
        async with get_events_subscriber(
            filter=EventFilter(
                event=EventNameFilter(
                    name=[
                        f"prefect.flow-run.{state.name.title()}"
                        for state in TERMINAL_STATES
                    ],
                )
            )
        ) as subscriber:
            consumer_started.set()
            async for event in subscriber:
                try:
                    self.logger.debug(
                        f"Received event: {event.resource['prefect.resource.id']}"
                    )
                    flow_run_id = uuid.UUID(
                        event.resource["prefect.resource.id"].replace(
                            "prefect.flow-run.", ""
                        )
                    )

                    with self._observed_completed_flow_runs_lock:
                        # Cache the flow run ID for a short period of time to avoid
                        # unnecessary waits
                        self._observed_completed_flow_runs[flow_run_id] = True
                    with self._completion_events_lock:
                        # Set the event for the flow run ID if it is in the cache
                        # so the waiter can wake up the waiting coroutine
                        if flow_run_id in self._completion_events:
                            self._completion_events[flow_run_id].set()
                        if flow_run_id in self._completion_callbacks:
                            self._completion_callbacks[flow_run_id]()
                except Exception as exc:
                    self.logger.error(f"Error processing event: {exc}")

    def stop(self) -> None:
        """
        Stop the FlowRunWaiter service.
        """
        self.logger.debug("Stopping FlowRunWaiter")
        if self._consumer_task:
            self._consumer_task.cancel()
            self._consumer_task = None
        self.__class__._instance = None
        self._started = False

    @classmethod
    async def wait_for_flow_run(
        cls, flow_run_id: uuid.UUID, timeout: float | None = None
    ) -> None:
        """
        Wait for a flow run to finish.

        Note this relies on a websocket connection to receive events from the server
        and will not work with an ephemeral server.

        Args:
            flow_run_id: The ID of the flow run to wait for.
            timeout: The maximum time to wait for the flow run to
                finish. Defaults to None.
        """
        instance = cls.instance()
        with instance._observed_completed_flow_runs_lock:
            if flow_run_id in instance._observed_completed_flow_runs:
                return

        # Need to create event in loop thread to ensure it can be set
        # from the loop thread
        finished_event = await from_async.wait_for_call_in_loop_thread(
            create_call(asyncio.Event)
        )
        with instance._completion_events_lock:
            # Cache the event for the flow run ID so the consumer can set it
            # when the event is received
            instance._completion_events[flow_run_id] = finished_event

        try:
            # Now check one more time whether the flow run arrived before we start to
            # wait on it, in case it came in while we were setting up the event above.
            with instance._observed_completed_flow_runs_lock:
                if flow_run_id in instance._observed_completed_flow_runs:
                    return

            with anyio.move_on_after(delay=timeout):
                await from_async.wait_for_call_in_loop_thread(
                    create_call(finished_event.wait)
                )
        finally:
            with instance._completion_events_lock:
                # Remove the event from the cache after it has been waited on
                instance._completion_events.pop(flow_run_id, None)

    @classmethod
    def add_done_callback(
        cls, flow_run_id: uuid.UUID, callback: Callable[[], None]
    ) -> None:
        """
        Add a callback to be called when a flow run finishes.

        Args:
            flow_run_id: The ID of the flow run to wait for.
            callback: The callback to call when the flow run finishes.
        """
        instance = cls.instance()
        with instance._observed_completed_flow_runs_lock:
            if flow_run_id in instance._observed_completed_flow_runs:
                callback()
                return

        with instance._completion_events_lock:
            # Cache the event for the flow run ID so the consumer can set it
            # when the event is received
            instance._completion_callbacks[flow_run_id] = callback

    @classmethod
    def instance(cls) -> Self:
        """
        Get the singleton instance of FlowRunWaiter.
        """
        with cls._instance_lock:
            if cls._instance is None:
                cls._instance = cls._new_instance()
            return cls._instance

    @classmethod
    def _new_instance(cls):
        instance = cls()

        if threading.get_ident() == get_global_loop().thread.ident:
            instance.start()
        else:
            from_sync.call_soon_in_loop_thread(create_call(instance.start)).result()

        return instance



from __future__ import annotations

import inspect
import os
import socket
import threading
import uuid
from functools import partial
from operator import methodcaller
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Callable,
    ClassVar,
    Optional,
    TypeVar,
    Union,
)
from uuid import UUID

from cachetools import LRUCache
from pydantic import (
    BaseModel,
    ConfigDict,
    Discriminator,
    Field,
    Tag,
)
from typing_extensions import ParamSpec, Self

import prefect
from prefect._internal.compatibility.async_dispatch import async_dispatch
from prefect._result_records import R, ResultRecord, ResultRecordMetadata
from prefect.blocks.core import Block
from prefect.exceptions import (
    ConfigurationError,
    MissingContextError,
)
from prefect.filesystems import (
    LocalFileSystem,
    NullFileSystem,
    WritableFileSystem,
)
from prefect.locking.protocol import LockManager
from prefect.logging import get_logger
from prefect.serializers import Serializer
from prefect.settings.context import get_current_settings
from prefect.types import DateTime
from prefect.utilities.annotations import NotSet
from prefect.utilities.asyncutils import sync_compatible

if TYPE_CHECKING:
    import logging

    from prefect import Flow, Task
    from prefect.transactions import IsolationLevel


ResultStorage = Union[WritableFileSystem, str]
ResultSerializer = Union[Serializer, str]
LITERAL_TYPES: set[type] = {type(None), bool, UUID}


def DEFAULT_STORAGE_KEY_FN() -> str:
    return uuid.uuid4().hex


logger: "logging.Logger" = get_logger("results")
P = ParamSpec("P")

_default_storages: dict[tuple[str, str], WritableFileSystem] = {}


async def aget_default_result_storage() -> WritableFileSystem:
    """
    Generate a default file system for result storage.
    """
    settings = get_current_settings()
    default_block = settings.results.default_storage_block
    basepath = settings.results.local_storage_path

    cache_key = (str(default_block), str(basepath))

    if cache_key in _default_storages:
        return _default_storages[cache_key]

    if default_block is not None:
        storage = await aresolve_result_storage(default_block)
    else:
        # Use the local file system
        storage = LocalFileSystem(basepath=str(basepath))

    _default_storages[cache_key] = storage
    return storage


@async_dispatch(aget_default_result_storage)
def get_default_result_storage() -> WritableFileSystem:
    """
    Generate a default file system for result storage.
    """
    settings = get_current_settings()
    default_block = settings.results.default_storage_block
    basepath = settings.results.local_storage_path

    cache_key = (str(default_block), str(basepath))

    if cache_key in _default_storages:
        return _default_storages[cache_key]

    if default_block is not None:
        storage = resolve_result_storage(default_block, _sync=True)
        if TYPE_CHECKING:
            assert isinstance(storage, WritableFileSystem)
    else:
        # Use the local file system
        storage = LocalFileSystem(basepath=str(basepath))

    _default_storages[cache_key] = storage
    return storage


async def aresolve_result_storage(
    result_storage: ResultStorage | UUID | Path,
) -> WritableFileSystem:
    """
    Resolve one of the valid `ResultStorage` input types into a saved block
    document id and an instance of the block.
    """
    from prefect.client.orchestration import get_client

    client = get_client()
    storage_block: WritableFileSystem
    if isinstance(result_storage, Block):
        storage_block = result_storage
    elif isinstance(result_storage, Path):
        storage_block = LocalFileSystem(basepath=str(result_storage))
    elif isinstance(result_storage, str):
        block = await Block.aload(result_storage, client=client)
        if TYPE_CHECKING:
            assert isinstance(block, WritableFileSystem)
        storage_block = block
    elif isinstance(result_storage, UUID):  # pyright: ignore[reportUnnecessaryIsInstance]
        block_document = await client.read_block_document(result_storage)
        from_block_document = methodcaller("_from_block_document", block_document)
        block = from_block_document(Block)
        if TYPE_CHECKING:
            assert isinstance(block, WritableFileSystem)
        storage_block = block
    else:
        raise TypeError(
            "Result storage must be one of the following types: 'UUID', 'Block', "
            f"'str'. Got unsupported type {type(result_storage).__name__!r}."
        )

    return storage_block


@async_dispatch(aresolve_result_storage)
def resolve_result_storage(
    result_storage: ResultStorage | UUID | Path,
) -> WritableFileSystem:
    """
    Resolve one of the valid `ResultStorage` input types into a saved block
    document id and an instance of the block.
    """
    from prefect.client.orchestration import get_client

    client = get_client(sync_client=True)
    storage_block: WritableFileSystem
    if isinstance(result_storage, Block):
        storage_block = result_storage
    elif isinstance(result_storage, Path):
        storage_block = LocalFileSystem(basepath=str(result_storage))
    elif isinstance(result_storage, str):
        block = Block.load(result_storage, _sync=True)
        if TYPE_CHECKING:
            assert isinstance(block, WritableFileSystem)
        storage_block = block
    elif isinstance(result_storage, UUID):  # pyright: ignore[reportUnnecessaryIsInstance]
        block_document = client.read_block_document(result_storage)
        from_block_document = methodcaller("_from_block_document", block_document)
        block = from_block_document(Block)
        if TYPE_CHECKING:
            assert isinstance(block, WritableFileSystem)
        storage_block = block
    else:
        raise TypeError(
            "Result storage must be one of the following types: 'UUID', 'Block', "
            f"'str'. Got unsupported type {type(result_storage).__name__!r}."
        )
    return storage_block


def resolve_serializer(serializer: ResultSerializer) -> Serializer:
    """
    Resolve one of the valid `ResultSerializer` input types into a serializer
    instance.
    """
    if isinstance(serializer, Serializer):
        return serializer
    elif isinstance(serializer, str):  # pyright: ignore[reportUnnecessaryIsInstance]
        return Serializer(type=serializer)
    else:
        raise TypeError(
            "Result serializer must be one of the following types: 'Serializer', "
            f"'str'. Got unsupported type {type(serializer).__name__!r}."
        )


async def get_or_create_default_task_scheduling_storage() -> ResultStorage:
    """
    Generate a default file system for background task parameter/result storage.
    """
    settings = get_current_settings()
    default_block = settings.tasks.scheduling.default_storage_block

    if default_block is not None:
        block = await Block.aload(default_block)
        if TYPE_CHECKING:
            assert isinstance(block, WritableFileSystem)
        return block

    # otherwise, use the local file system
    basepath = settings.results.local_storage_path
    return LocalFileSystem(basepath=str(basepath))


def get_default_result_serializer() -> Serializer:
    """
    Generate a default file system for result storage.
    """
    settings = get_current_settings()
    return resolve_serializer(settings.results.default_serializer)


def get_default_persist_setting() -> bool:
    """
    Return the default option for result persistence.
    """
    settings = get_current_settings()
    return settings.results.persist_by_default


def get_default_persist_setting_for_tasks() -> bool:
    """
    Return the default option for result persistence for tasks.
    """
    settings = get_current_settings()
    return (
        settings.tasks.default_persist_result
        if settings.tasks.default_persist_result is not None
        else settings.results.persist_by_default
    )


def should_persist_result() -> bool:
    """
    Return the default option for result persistence determined by the current run context.

    If there is no current run context, the value of `results.persist_by_default` on the
    current settings will be returned.
    """
    from prefect.context import FlowRunContext, TaskRunContext

    task_run_context = TaskRunContext.get()
    if task_run_context is not None:
        return task_run_context.persist_result
    flow_run_context = FlowRunContext.get()
    if flow_run_context is not None:
        return flow_run_context.persist_result

    return get_default_persist_setting()


def _format_user_supplied_storage_key(key: str) -> str:
    # Note here we are pinning to task runs since flow runs do not support storage keys
    # yet; we'll need to split logic in the future or have two separate functions
    runtime_vars = {key: getattr(prefect.runtime, key) for key in dir(prefect.runtime)}
    return key.format(**runtime_vars, parameters=prefect.runtime.task_run.parameters)


async def _call_explicitly_async_block_method(
    block: WritableFileSystem | NullFileSystem,
    method: str,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    """
    TODO: remove this once we have explicit async methods on all storage blocks

    see https://github.com/PrefectHQ/prefect/issues/15008
    """
    if hasattr(block, f"a{method}"):  # explicit async method
        return await getattr(block, f"a{method}")(*args, **kwargs)
    elif hasattr(getattr(block, method, None), "aio"):  # sync_compatible
        return await getattr(block, method).aio(block, *args, **kwargs)
    else:  # should not happen in prefect, but users can override impls
        maybe_coro = getattr(block, method)(*args, **kwargs)
        if inspect.isawaitable(maybe_coro):
            return await maybe_coro
        else:
            return maybe_coro


T = TypeVar("T")


def default_cache() -> LRUCache[str, "ResultRecord[Any]"]:
    return LRUCache(maxsize=1000)


def result_storage_discriminator(x: Any) -> str:
    if isinstance(x, dict):
        if "block_type_slug" in x:
            return "WritableFileSystem"
        else:
            return "NullFileSystem"
    if isinstance(x, WritableFileSystem):
        return "WritableFileSystem"
    if isinstance(x, NullFileSystem):
        return "NullFileSystem"
    return "None"


class ResultStore(BaseModel):
    """
    Manages the storage and retrieval of results.

    Attributes:
        result_storage: The storage for result records. If not provided, the default
            result storage will be used.
        metadata_storage: The storage for result record metadata. If not provided,
            the metadata will be stored alongside the results.
        lock_manager: The lock manager to use for locking result records. If not provided,
            the store cannot be used in transactions with the SERIALIZABLE isolation level.
        cache_result_in_memory: Whether to cache results in memory.
        serializer: The serializer to use for results.
        storage_key_fn: The function to generate storage keys.
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    result_storage: Optional[WritableFileSystem] = Field(default=None)
    metadata_storage: Annotated[
        Union[
            Annotated[WritableFileSystem, Tag("WritableFileSystem")],
            Annotated[NullFileSystem, Tag("NullFileSystem")],
            Annotated[None, Tag("None")],
        ],
        Discriminator(result_storage_discriminator),
    ] = Field(default=None)
    lock_manager: Optional[LockManager] = Field(default=None)
    cache_result_in_memory: bool = Field(default=True)
    serializer: Serializer = Field(default_factory=get_default_result_serializer)
    storage_key_fn: Callable[[], str] = Field(default=DEFAULT_STORAGE_KEY_FN)
    cache: LRUCache[str, "ResultRecord[Any]"] = Field(default_factory=default_cache)

    @property
    def result_storage_block_id(self) -> UUID | None:
        if self.result_storage is None:
            return None
        return getattr(self.result_storage, "_block_document_id", None)

    @classmethod
    async def _from_metadata(cls, metadata: ResultRecordMetadata) -> "ResultRecord[R]":
        """
        Create a result record from metadata.

        Will use the result record metadata to fetch data via a result store.

        Args:
            metadata: The metadata to create the result record from.

        Returns:
            ResultRecord: The result record.
        """
        if metadata.storage_block_id is None:
            storage_block = None
        else:
            storage_block = await aresolve_result_storage(metadata.storage_block_id)
        store = cls(result_storage=storage_block, serializer=metadata.serializer)
        if metadata.storage_key is None:
            raise ValueError(
                "storage_key is required to hydrate a result record from metadata"
            )
        result = await store.aread(metadata.storage_key)
        return result

    @sync_compatible
    async def update_for_flow(self, flow: "Flow[..., Any]") -> Self:
        """
        Create a new result store for a flow with updated settings.

        Args:
            flow: The flow to update the result store for.

        Returns:
            An updated result store.
        """
        update: dict[str, Any] = {}
        update["cache_result_in_memory"] = flow.cache_result_in_memory
        if flow.result_storage is not None:
            update["result_storage"] = await aresolve_result_storage(
                flow.result_storage
            )
        if flow.result_serializer is not None:
            update["serializer"] = resolve_serializer(flow.result_serializer)
        if self.result_storage is None and update.get("result_storage") is None:
            update["result_storage"] = await aget_default_result_storage()
        update["metadata_storage"] = NullFileSystem()
        return self.model_copy(update=update)

    @sync_compatible
    async def update_for_task(self: Self, task: "Task[P, R]") -> Self:
        """
        Create a new result store for a task.

        Args:
            task: The task to update the result store for.

        Returns:
            An updated result store.
        """
        from prefect.transactions import get_transaction

        update: dict[str, Any] = {}
        update["cache_result_in_memory"] = task.cache_result_in_memory
        if task.result_storage is not None:
            update["result_storage"] = await aresolve_result_storage(
                task.result_storage
            )
        if task.result_serializer is not None:
            update["serializer"] = resolve_serializer(task.result_serializer)
        if task.result_storage_key is not None:
            update["storage_key_fn"] = partial(
                _format_user_supplied_storage_key, task.result_storage_key
            )

        # use the lock manager from a parent transaction if it exists
        if (current_txn := get_transaction()) and isinstance(
            current_txn.store, ResultStore
        ):
            update["lock_manager"] = current_txn.store.lock_manager

        from prefect.cache_policies import CachePolicy

        if isinstance(task.cache_policy, CachePolicy):
            if task.cache_policy.key_storage is not None:
                storage = task.cache_policy.key_storage
                if isinstance(storage, str) and not len(storage.split("/")) == 2:
                    storage = Path(storage)
                update["metadata_storage"] = await aresolve_result_storage(storage)
            # if the cache policy has a lock manager, it takes precedence over the parent transaction
            if task.cache_policy.lock_manager is not None:
                update["lock_manager"] = task.cache_policy.lock_manager

        if self.result_storage is None and update.get("result_storage") is None:
            update["result_storage"] = await aget_default_result_storage()
        if (
            isinstance(self.metadata_storage, NullFileSystem)
            and update.get("metadata_storage", NotSet) is NotSet
        ):
            update["metadata_storage"] = None
        return self.model_copy(update=update)

    @staticmethod
    def generate_default_holder() -> str:
        """
        Generate a default holder string using hostname, PID, and thread ID.

        Returns:
            str: A unique identifier string.
        """
        hostname = socket.gethostname()
        pid = os.getpid()
        thread_name = threading.current_thread().name
        thread_id = threading.get_ident()
        return f"{hostname}:{pid}:{thread_id}:{thread_name}"

    @sync_compatible
    async def _exists(self, key: str) -> bool:
        """
        Check if a result record exists in storage.

        Args:
            key: The key to check for the existence of a result record.

        Returns:
            bool: True if the result record exists, False otherwise.
        """
        if self.metadata_storage is not None:
            # TODO: Add an `exists` method to commonly used storage blocks
            # so the entire payload doesn't need to be read
            try:
                metadata_content = await _call_explicitly_async_block_method(
                    self.metadata_storage, "read_path", (key,), {}
                )
                if metadata_content is None:
                    return False
                metadata = ResultRecordMetadata.load_bytes(metadata_content)

            except Exception:
                return False
        else:
            try:
                content = await _call_explicitly_async_block_method(
                    self.result_storage, "read_path", (key,), {}
                )
                if content is None:
                    return False
                record: ResultRecord[Any] = ResultRecord.deserialize(content)
                metadata = record.metadata
            except Exception:
                return False

        if metadata.expiration:
            # if the result has an expiration,
            # check if it is still in the future
            exists = metadata.expiration > DateTime.now("utc")
        else:
            exists = True
        return exists

    def exists(self, key: str) -> bool:
        """
        Check if a result record exists in storage.

        Args:
            key: The key to check for the existence of a result record.

        Returns:
            bool: True if the result record exists, False otherwise.
        """
        return self._exists(key=key, _sync=True)

    async def aexists(self, key: str) -> bool:
        """
        Check if a result record exists in storage.

        Args:
            key: The key to check for the existence of a result record.

        Returns:
            bool: True if the result record exists, False otherwise.
        """
        return await self._exists(key=key, _sync=False)

    def _resolved_key_path(self, key: str) -> str:
        if self.result_storage_block_id is None and (
            _resolve_path := getattr(self.result_storage, "_resolve_path", None)
        ):
            path_key = _resolve_path(key)
            if path_key is not None:
                return str(_resolve_path(key))
            else:
                return key
        return key

    @sync_compatible
    async def _read(self, key: str, holder: str) -> "ResultRecord[Any]":
        """
        Read a result record from storage.

        This is the internal implementation. Use `read` or `aread` for synchronous and
        asynchronous result reading respectively.

        Args:
            key: The key to read the result record from.
            holder: The holder of the lock if a lock was set on the record.

        Returns:
            A result record.
        """

        from prefect._experimental.lineage import emit_result_read_event

        if self.lock_manager is not None and not self.is_lock_holder(key, holder):
            await self.await_for_lock(key)

        resolved_key_path = self._resolved_key_path(key)

        if resolved_key_path in self.cache:
            cached_result: ResultRecord[Any] = self.cache[resolved_key_path]
            await emit_result_read_event(self, resolved_key_path, cached=True)
            return cached_result

        if self.result_storage is None:
            self.result_storage = await aget_default_result_storage()

        if self.metadata_storage is not None:
            metadata_content = await _call_explicitly_async_block_method(
                self.metadata_storage,
                "read_path",
                (key,),
                {},
            )
            metadata = ResultRecordMetadata.load_bytes(metadata_content)
            assert metadata.storage_key is not None, (
                "Did not find storage key in metadata"
            )
            result_content = await _call_explicitly_async_block_method(
                self.result_storage,
                "read_path",
                (metadata.storage_key,),
                {},
            )
            result_record: ResultRecord[Any] = (
                ResultRecord.deserialize_from_result_and_metadata(
                    result=result_content, metadata=metadata_content
                )
            )
            await emit_result_read_event(self, resolved_key_path)
        else:
            content = await _call_explicitly_async_block_method(
                self.result_storage,
                "read_path",
                (key,),
                {},
            )
            result_record: ResultRecord[Any] = ResultRecord.deserialize(
                content, backup_serializer=self.serializer
            )
            await emit_result_read_event(self, resolved_key_path)

        if self.cache_result_in_memory:
            self.cache[resolved_key_path] = result_record
        return result_record

    def read(
        self,
        key: str,
        holder: str | None = None,
    ) -> "ResultRecord[Any]":
        """
        Read a result record from storage.

        Args:
            key: The key to read the result record from.
            holder: The holder of the lock if a lock was set on the record.

        Returns:
            A result record.
        """
        holder = holder or self.generate_default_holder()
        return self._read(key=key, holder=holder, _sync=True)

    async def aread(
        self,
        key: str,
        holder: str | None = None,
    ) -> "ResultRecord[Any]":
        """
        Read a result record from storage.

        Args:
            key: The key to read the result record from.
            holder: The holder of the lock if a lock was set on the record.

        Returns:
            A result record.
        """
        holder = holder or self.generate_default_holder()
        return await self._read(key=key, holder=holder, _sync=False)

    def create_result_record(
        self,
        obj: Any,
        key: str | None = None,
        expiration: DateTime | None = None,
    ) -> "ResultRecord[Any]":
        """
        Create a result record.

        Args:
            key: The key to create the result record for.
            obj: The object to create the result record for.
            expiration: The expiration time for the result record.
        """
        key = key or self.storage_key_fn()

        if self.result_storage is None:
            self.result_storage = get_default_result_storage(_sync=True)
            if TYPE_CHECKING:
                assert isinstance(self.result_storage, WritableFileSystem)

        if self.result_storage_block_id is None:
            if _resolve_path := getattr(self.result_storage, "_resolve_path", None):
                path_key = _resolve_path(key)
                if path_key is not None:
                    key = str(_resolve_path(key))

        return ResultRecord(
            result=obj,
            metadata=ResultRecordMetadata(
                serializer=self.serializer,
                expiration=expiration,
                storage_key=key,
                storage_block_id=self.result_storage_block_id,
            ),
        )

    def write(
        self,
        obj: Any,
        key: str | None = None,
        expiration: DateTime | None = None,
        holder: str | None = None,
    ) -> None:
        """
        Write a result to storage.

        Handles the creation of a `ResultRecord` and its serialization to storage.

        Args:
            key: The key to write the result record to.
            obj: The object to write to storage.
            expiration: The expiration time for the result record.
            holder: The holder of the lock if a lock was set on the record.
        """
        holder = holder or self.generate_default_holder()
        result_record = self.create_result_record(
            key=key, obj=obj, expiration=expiration
        )
        return self.persist_result_record(
            result_record=result_record,
            holder=holder,
        )

    async def awrite(
        self,
        obj: Any,
        key: str | None = None,
        expiration: DateTime | None = None,
        holder: str | None = None,
    ) -> None:
        """
        Write a result to storage.

        Args:
            key: The key to write the result record to.
            obj: The object to write to storage.
            expiration: The expiration time for the result record.
            holder: The holder of the lock if a lock was set on the record.
        """
        holder = holder or self.generate_default_holder()
        return await self.apersist_result_record(
            result_record=self.create_result_record(
                key=key, obj=obj, expiration=expiration
            ),
            holder=holder,
        )

    @sync_compatible
    async def _persist_result_record(
        self, result_record: "ResultRecord[Any]", holder: str
    ) -> None:
        """
        Persist a result record to storage.

        Args:
            result_record: The result record to persist.
            holder: The holder of the lock if a lock was set on the record.
        """
        assert result_record.metadata.storage_key is not None, (
            "Storage key is required on result record"
        )

        from prefect._experimental.lineage import emit_result_write_event

        key = result_record.metadata.storage_key
        if result_record.metadata.storage_block_id is None:
            basepath = (
                _resolve_path("")
                if (
                    _resolve_path := getattr(self.result_storage, "_resolve_path", None)
                )
                else Path(".").resolve()
            )
            base_key = key if basepath is None else str(Path(key).relative_to(basepath))
        else:
            base_key = key
        if (
            self.lock_manager is not None
            and self.is_locked(base_key)
            and not self.is_lock_holder(base_key, holder)
        ):
            raise RuntimeError(
                f"Cannot write to result record with key {base_key} because it is locked by "
                f"another holder."
            )
        if self.result_storage is None:
            self.result_storage = await aget_default_result_storage()

        # If metadata storage is configured, write result and metadata separately
        if self.metadata_storage is not None:
            await _call_explicitly_async_block_method(
                self.result_storage,
                "write_path",
                (result_record.metadata.storage_key,),
                {"content": result_record.serialize_result()},
            )
            await _call_explicitly_async_block_method(
                self.metadata_storage,
                "write_path",
                (base_key,),
                {"content": result_record.serialize_metadata()},
            )
            await emit_result_write_event(self, result_record.metadata.storage_key)
        # Otherwise, write the result metadata and result together
        else:
            await _call_explicitly_async_block_method(
                self.result_storage,
                "write_path",
                (result_record.metadata.storage_key,),
                {"content": result_record.serialize()},
            )
            await emit_result_write_event(self, result_record.metadata.storage_key)
        if self.cache_result_in_memory:
            self.cache[key] = result_record

    def persist_result_record(
        self, result_record: "ResultRecord[Any]", holder: str | None = None
    ) -> None:
        """
        Persist a result record to storage.

        Args:
            result_record: The result record to persist.
        """
        holder = holder or self.generate_default_holder()
        return self._persist_result_record(
            result_record=result_record, holder=holder, _sync=True
        )

    async def apersist_result_record(
        self, result_record: "ResultRecord[Any]", holder: str | None = None
    ) -> None:
        """
        Persist a result record to storage.

        Args:
            result_record: The result record to persist.
        """
        holder = holder or self.generate_default_holder()
        return await self._persist_result_record(
            result_record=result_record, holder=holder, _sync=False
        )

    def supports_isolation_level(self, level: "IsolationLevel") -> bool:
        """
        Check if the result store supports a given isolation level.

        Args:
            level: The isolation level to check.

        Returns:
            bool: True if the isolation level is supported, False otherwise.
        """
        from prefect.transactions import IsolationLevel

        if level == IsolationLevel.READ_COMMITTED:
            return True
        elif level == IsolationLevel.SERIALIZABLE:
            return self.lock_manager is not None
        else:
            raise ValueError(f"Unsupported isolation level: {level}")

    def acquire_lock(
        self, key: str, holder: str | None = None, timeout: float | None = None
    ) -> bool:
        """
        Acquire a lock for a result record.

        Args:
            key: The key to acquire the lock for.
            holder: The holder of the lock. If not provided, a default holder based on the
                current host, process, and thread will be used.
            timeout: The timeout for the lock.

        Returns:
            bool: True if the lock was successfully acquired; False otherwise.
        """
        holder = holder or self.generate_default_holder()
        if self.lock_manager is None:
            raise ConfigurationError(
                "Result store is not configured with a lock manager. Please set"
                " a lock manager when creating the result store to enable locking."
            )
        return self.lock_manager.acquire_lock(key, holder, timeout)

    async def aacquire_lock(
        self, key: str, holder: str | None = None, timeout: float | None = None
    ) -> bool:
        """
        Acquire a lock for a result record.

        Args:
            key: The key to acquire the lock for.
            holder: The holder of the lock. If not provided, a default holder based on the
                current host, process, and thread will be used.
            timeout: The timeout for the lock.

        Returns:
            bool: True if the lock was successfully acquired; False otherwise.
        """
        holder = holder or self.generate_default_holder()
        if self.lock_manager is None:
            raise ConfigurationError(
                "Result store is not configured with a lock manager. Please set"
                " a lock manager when creating the result store to enable locking."
            )

        return await self.lock_manager.aacquire_lock(key, holder, timeout)

    def release_lock(self, key: str, holder: str | None = None) -> None:
        """
        Release a lock for a result record.

        Args:
            key: The key to release the lock for.
            holder: The holder of the lock. Must match the holder that acquired the lock.
                If not provided, a default holder based on the current host, process, and
                thread will be used.
        """
        holder = holder or self.generate_default_holder()
        if self.lock_manager is None:
            raise ConfigurationError(
                "Result store is not configured with a lock manager. Please set"
                " a lock manager when creating the result store to enable locking."
            )
        return self.lock_manager.release_lock(key, holder)

    def is_locked(self, key: str) -> bool:
        """
        Check if a result record is locked.
        """
        if self.lock_manager is None:
            raise ConfigurationError(
                "Result store is not configured with a lock manager. Please set"
                " a lock manager when creating the result store to enable locking."
            )
        return self.lock_manager.is_locked(key)

    def is_lock_holder(self, key: str, holder: str | None = None) -> bool:
        """
        Check if the current holder is the lock holder for the result record.

        Args:
            key: The key to check the lock for.
            holder: The holder of the lock. If not provided, a default holder based on the
                current host, process, and thread will be used.

        Returns:
            bool: True if the current holder is the lock holder; False otherwise.
        """
        holder = holder or self.generate_default_holder()
        if self.lock_manager is None:
            raise ConfigurationError(
                "Result store is not configured with a lock manager. Please set"
                " a lock manager when creating the result store to enable locking."
            )
        return self.lock_manager.is_lock_holder(key, holder)

    def wait_for_lock(self, key: str, timeout: float | None = None) -> bool:
        """
        Wait for the corresponding transaction record to become free.
        """
        if self.lock_manager is None:
            raise ConfigurationError(
                "Result store is not configured with a lock manager. Please set"
                " a lock manager when creating the result store to enable locking."
            )
        return self.lock_manager.wait_for_lock(key, timeout)

    async def await_for_lock(self, key: str, timeout: float | None = None) -> bool:
        """
        Wait for the corresponding transaction record to become free.
        """
        if self.lock_manager is None:
            raise ConfigurationError(
                "Result store is not configured with a lock manager. Please set"
                " a lock manager when creating the result store to enable locking."
            )
        return await self.lock_manager.await_for_lock(key, timeout)

    # TODO: These two methods need to find a new home

    @sync_compatible
    async def store_parameters(self, identifier: UUID, parameters: dict[str, Any]):
        record = ResultRecord(
            result=parameters,
            metadata=ResultRecordMetadata(
                serializer=self.serializer, storage_key=str(identifier)
            ),
        )

        await _call_explicitly_async_block_method(
            self.result_storage,
            "write_path",
            (f"parameters/{identifier}",),
            {"content": record.serialize()},
        )

    @sync_compatible
    async def read_parameters(self, identifier: UUID) -> dict[str, Any]:
        if self.result_storage is None:
            raise ValueError(
                "Result store is not configured - must have a result storage block to read parameters"
            )
        record: ResultRecord[Any] = ResultRecord.deserialize(
            await _call_explicitly_async_block_method(
                self.result_storage,
                "read_path",
                (f"parameters/{identifier}",),
                {},
            )
        )
        return record.result


def get_result_store() -> ResultStore:
    """
    Get the current result store.
    """
    from prefect.context import get_run_context

    try:
        run_context = get_run_context()
    except MissingContextError:
        result_store = ResultStore()
    else:
        result_store = run_context.result_store
    return result_store



from typing import TYPE_CHECKING, Sequence
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.asyncio import AsyncSession

from prefect.server.database import PrefectDBInterface, db_injector
from prefect.server.events.schemas.automations import CompositeTrigger, Firing
from prefect.types._datetime import DateTime, now

if TYPE_CHECKING:
    from prefect.server.database.orm_models import ORMCompositeTriggerChildFiring


@db_injector
async def upsert_child_firing(
    db: PrefectDBInterface,
    session: AsyncSession,
    firing: Firing,
):
    automation_id = firing.trigger.automation.id
    parent_trigger_id = firing.trigger.parent.id
    child_trigger_id = firing.trigger.id

    upsert = (
        postgresql.insert(db.CompositeTriggerChildFiring)
        .values(
            automation_id=automation_id,
            parent_trigger_id=parent_trigger_id,
            child_trigger_id=child_trigger_id,
            child_firing_id=firing.id,
            child_fired_at=firing.triggered,
            child_firing=firing.model_dump(),
        )
        .on_conflict_do_update(
            index_elements=[
                db.CompositeTriggerChildFiring.automation_id,
                db.CompositeTriggerChildFiring.parent_trigger_id,
                db.CompositeTriggerChildFiring.child_trigger_id,
            ],
            set_=dict(
                child_firing_id=firing.id,
                child_fired_at=firing.triggered,
                child_firing=firing.model_dump(),
                updated=now("UTC"),
            ),
        )
    )

    await session.execute(upsert)

    result = await session.execute(
        sa.select(db.CompositeTriggerChildFiring).filter(
            db.CompositeTriggerChildFiring.automation_id == automation_id,
            db.CompositeTriggerChildFiring.parent_trigger_id == parent_trigger_id,
            db.CompositeTriggerChildFiring.child_trigger_id == child_trigger_id,
        )
    )

    return result.scalars().one()


@db_injector
async def get_child_firings(
    db: PrefectDBInterface,
    session: AsyncSession,
    trigger: CompositeTrigger,
) -> Sequence["ORMCompositeTriggerChildFiring"]:
    result = await session.execute(
        sa.select(db.CompositeTriggerChildFiring).filter(
            db.CompositeTriggerChildFiring.automation_id == trigger.automation.id,
            db.CompositeTriggerChildFiring.parent_trigger_id == trigger.id,
            db.CompositeTriggerChildFiring.child_trigger_id.in_(
                trigger.child_trigger_ids
            ),
        )
    )

    return result.scalars().unique().all()


@db_injector
async def clear_old_child_firings(
    db: PrefectDBInterface,
    session: AsyncSession,
    trigger: CompositeTrigger,
    fired_before: DateTime,
) -> None:
    await session.execute(
        sa.delete(db.CompositeTriggerChildFiring).filter(
            db.CompositeTriggerChildFiring.automation_id == trigger.automation.id,
            db.CompositeTriggerChildFiring.parent_trigger_id == trigger.id,
            db.CompositeTriggerChildFiring.child_fired_at < fired_before,
        )
    )


@db_injector
async def clear_child_firings(
    db: PrefectDBInterface,
    session: AsyncSession,
    trigger: CompositeTrigger,
    firing_ids: Sequence[UUID],
) -> None:
    await session.execute(
        sa.delete(db.CompositeTriggerChildFiring).filter(
            db.CompositeTriggerChildFiring.automation_id == trigger.automation.id,
            db.CompositeTriggerChildFiring.parent_trigger_id == trigger.id,
            db.CompositeTriggerChildFiring.child_firing_id.in_(firing_ids),
        )
    )



from collections.abc import Sequence
from typing import Any, Protocol, TypeVar

from opentelemetry.exporter.otlp.proto.http._log_exporter import OTLPLogExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk._logs import LogData
from opentelemetry.sdk._logs.export import LogExporter
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult

from prefect._internal.concurrency.services import BatchedQueueService

BatchItem = TypeVar("BatchItem", ReadableSpan, LogData)
T_contra = TypeVar("T_contra", contravariant=True)


class OTLPExporter(Protocol[T_contra]):
    def export(self, __items: Sequence[T_contra]) -> Any: ...

    def shutdown(self) -> Any: ...


class BaseQueueingExporter(BatchedQueueService[BatchItem]):
    _max_batch_size = 512
    _min_interval = 2.0

    def __init__(self, otlp_exporter: OTLPExporter[BatchItem]) -> None:
        super().__init__()
        self._otlp_exporter = otlp_exporter

    async def _handle_batch(self, items: list[BatchItem]) -> None:
        try:
            self._otlp_exporter.export(items)
        except Exception as e:
            self._logger.exception(f"Failed to export batch: {e}")
            raise

    def shutdown(self) -> None:
        if self._stopped:
            return

        self.drain()
        self._otlp_exporter.shutdown()


class QueueingSpanExporter(BaseQueueingExporter[ReadableSpan], SpanExporter):
    _otlp_exporter: OTLPSpanExporter

    def __init__(self, endpoint: str, headers: tuple[tuple[str, str]]):
        super().__init__(OTLPSpanExporter(endpoint=endpoint, headers=dict(headers)))

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for item in spans:
            if self._stopped:
                break
            self.send(item)
        return SpanExportResult.SUCCESS


class QueueingLogExporter(BaseQueueingExporter[LogData], LogExporter):
    _otlp_exporter: OTLPLogExporter

    def __init__(self, endpoint: str, headers: tuple[tuple[str, str]]) -> None:
        super().__init__(OTLPLogExporter(endpoint=endpoint, headers=dict(headers)))

    def export(self, batch: Sequence[LogData]) -> None:
        for item in batch:
            if self._stopped:
                break
            self.send(item)



"""
Functions for interacting with block schema ORM objects.
Intended for internal use by the Prefect REST API.
"""

import json
from copy import copy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession

from prefect.server import schemas
from prefect.server.database import PrefectDBInterface, db_injector, orm_models
from prefect.server.models.block_types import read_block_type_by_slug
from prefect.server.schemas.actions import BlockSchemaCreate
from prefect.server.schemas.core import BlockSchema, BlockSchemaReference

if TYPE_CHECKING:
    from prefect.client.schemas.actions import (
        BlockSchemaCreate as ClientBlockSchemaCreate,
    )
    from prefect.client.schemas.objects import BlockSchema as ClientBlockSchema


class MissingBlockTypeException(Exception):
    """Raised when the block type corresponding to a block schema cannot be found"""


@db_injector
async def create_block_schema(
    db: PrefectDBInterface,
    session: AsyncSession,
    block_schema: Union[
        schemas.actions.BlockSchemaCreate,
        schemas.core.BlockSchema,
        "ClientBlockSchemaCreate",
        "ClientBlockSchema",
    ],
    override: bool = False,
    definitions: Optional[dict[str, Any]] = None,
) -> Union[BlockSchema, orm_models.BlockSchema]:
    """
    Create a new block schema.

    Args:
        session: A database session
        block_schema: a block schema object
        definitions: Definitions of fields from block schema fields
            attribute. Used when recursively creating nested block schemas

    Returns:
        block_schema: an ORM block schema model
    """
    from prefect.blocks.core import Block, _get_non_block_reference_definitions

    # We take a shortcut in many unit tests and in block registration to pass client
    # models directly to this function.  We will support this by converting them to
    # the appropriate server model.
    if not isinstance(block_schema, schemas.actions.BlockSchemaCreate):
        block_schema = schemas.actions.BlockSchemaCreate.model_validate(
            block_schema.model_dump(
                mode="json",
                exclude={"id", "created", "updated", "checksum", "block_type"},
            )
        )

    insert_values = block_schema.model_dump_for_orm(
        exclude_unset=False,
        exclude={"block_type", "id", "created", "updated"},
    )

    definitions = definitions or block_schema.fields.get("definitions")
    fields_for_checksum = insert_values["fields"]
    if definitions:
        # Ensure definitions are available if this is a nested schema
        # that is being registered
        fields_for_checksum["definitions"] = definitions
    checksum = Block._calculate_schema_checksum(fields_for_checksum)

    # Check for existing block schema based on calculated checksum
    existing_block_schema = await read_block_schema_by_checksum(
        session=session, checksum=checksum, version=block_schema.version
    )
    # Return existing block schema if it exists. Allows block schema creation to be called multiple
    # times for the same schema without errors.
    if existing_block_schema:
        return existing_block_schema

    insert_values["checksum"] = checksum

    if definitions:
        # Get non block definitions for saving to the DB.
        non_block_definitions = _get_non_block_reference_definitions(
            insert_values["fields"], definitions
        )
        if non_block_definitions:
            insert_values["fields"]["definitions"] = (
                _get_non_block_reference_definitions(
                    insert_values["fields"], definitions
                )
            )
        else:
            # Prevent storing definitions for blocks. Those are reconstructed on read.
            insert_values["fields"].pop("definitions", None)

    # Prevent saving block schema references in the block_schema table. They have
    # their own table.
    block_schema_references: Dict = insert_values["fields"].pop(
        "block_schema_references", {}
    )

    insert_stmt = db.queries.insert(db.BlockSchema).values(**insert_values)
    if override:
        insert_stmt = insert_stmt.on_conflict_do_update(
            index_elements=db.orm.block_schema_unique_upsert_columns,
            set_=insert_values,
        )
    await session.execute(insert_stmt)

    query = (
        sa.select(db.BlockSchema)
        .where(
            db.BlockSchema.checksum == insert_values["checksum"],
        )
        .order_by(db.BlockSchema.created.desc())
        .limit(1)
        .execution_options(populate_existing=True)
    )

    if block_schema.version is not None:
        query = query.where(db.BlockSchema.version == block_schema.version)

    result = await session.execute(query)
    created_block_schema = copy(result.scalar_one())

    await _register_nested_block_schemas(
        db,
        session=session,
        parent_block_schema_id=created_block_schema.id,
        block_schema_references=block_schema_references,
        base_fields=insert_values["fields"],
        definitions=definitions,
        override=override,
    )

    created_block_schema.fields["block_schema_references"] = block_schema_references
    if definitions is not None:
        created_block_schema.fields["definitions"] = definitions

    return created_block_schema


async def _register_nested_block_schemas(
    db: PrefectDBInterface,
    session: AsyncSession,
    parent_block_schema_id: UUID,
    block_schema_references: dict[str, Union[dict[str, str], List[dict[str, str]]]],
    base_fields: Dict,
    definitions: Optional[Dict],
    override: bool = False,
) -> None:
    """
    Iterates through each of the block schema references declared on the block schema.
    Attempts to register each of the nested block schemas if they have not already been
    registered. An error is thrown if the corresponding block type for a block schema
    has not been registered.

    Args:
        session: A database session.
        parent_block_schema_id: The ID of the parent block schema.
        block_schema_references: A dictionary containing the block schema references for
            the child block schemas of the parent block schema.
        base_fields: The field name and type declarations for the parent block schema.
        definitions: A dictionary of the field name and type declarations of each
            child block schema.
        override: Flag controlling if a block schema should updated in place.
    """
    for reference_name, reference_values in block_schema_references.items():
        # Operate on a list so that we can share the same code paths for union cases
        reference_values = (
            reference_values
            if isinstance(reference_values, list)
            else [reference_values]
        )
        for reference_values_entry in reference_values:
            # Check to make sure that associated block type exists
            reference_block_type = await read_block_type_by_slug(
                session=session,
                block_type_slug=reference_values_entry["block_type_slug"],
            )
            if reference_block_type is None:
                raise MissingBlockTypeException(
                    "Cannot create block schema because block type"
                    f" {reference_values_entry['block_type_slug']!r} was not found.Did"
                    " you forget to register the block type?"
                )

            reference_block_schema: Union[BlockSchema, orm_models.BlockSchema, None]

            # Checks to see if the visited block schema has been previously created
            reference_block_schema = await read_block_schema_by_checksum(
                session=session,
                checksum=reference_values_entry["block_schema_checksum"],
            )
            # Attempts to create block schema since it has not already been registered
            if reference_block_schema is None:
                if definitions is None:
                    raise ValueError(
                        "Unable to create nested block schema due to missing"
                        " definitions in root block schema fields"
                    )
                sub_block_schema_fields = _get_fields_for_child_schema(
                    db, definitions, base_fields, reference_name, reference_block_type
                )

                if sub_block_schema_fields is None:
                    raise ValueError(
                        "Unable to create nested block schema for block type"
                        f" {reference_block_type.name!r} due to missing definition."
                    )

                reference_block_schema = await create_block_schema(
                    session=session,
                    block_schema=BlockSchemaCreate(
                        fields=sub_block_schema_fields,
                        block_type_id=reference_block_type.id,
                    ),
                    override=override,
                    definitions=definitions,
                )
            # Create a block schema reference linking the nested block schema to its parent.
            await create_block_schema_reference(
                session=session,
                block_schema_reference=BlockSchemaReference(
                    parent_block_schema_id=parent_block_schema_id,
                    reference_block_schema_id=reference_block_schema.id,
                    name=reference_name,
                ),
            )


def _get_fields_for_child_schema(
    db: PrefectDBInterface,
    definitions: Dict,
    base_fields: Dict,
    reference_name: str,
    reference_block_type: orm_models.BlockType,
) -> dict[str, Any]:
    """
    Returns the field definitions for a child schema. The fields definitions are pulled from the provided `definitions`
    dictionary based on the information extracted from `base_fields` using the `reference_name`. `reference_block_type`
    is used to disambiguate fields that have a union type.
    """
    from prefect.blocks.core import _collect_nested_reference_strings

    spec_reference = base_fields["properties"][reference_name]
    sub_block_schema_fields = None
    reference_strings = _collect_nested_reference_strings(spec_reference)
    if len(reference_strings) == 1:
        sub_block_schema_fields = definitions.get(
            reference_strings[0].replace("#/definitions/", "")
        )
    else:
        for reference_string in reference_strings:
            definition_key = reference_string.replace("#/definitions/", "")
            potential_sub_block_schema_fields = definitions[definition_key]
            # Determines the definition to use when registering a child
            # block schema by verifying that the block type name stored in
            # the definition matches the name of the block type that we're
            # currently trying to register a block schema for.
            if (
                definitions[definition_key]["block_type_slug"]
                == reference_block_type.slug
            ):
                # Once we've found the matching definition, we no longer
                # need to iterate
                sub_block_schema_fields = potential_sub_block_schema_fields
                break
    return sub_block_schema_fields  # type: ignore


@db_injector
async def delete_block_schema(
    db: PrefectDBInterface, session: AsyncSession, block_schema_id: UUID
) -> bool:
    """
    Delete a block schema by id.

    Args:
        session: A database session
        block_schema_id: a block schema id

    Returns:
        bool: whether or not the block schema was deleted
    """

    result = await session.execute(
        delete(db.BlockSchema).where(db.BlockSchema.id == block_schema_id)
    )
    return result.rowcount > 0


@db_injector
async def read_block_schema(
    db: PrefectDBInterface,
    session: AsyncSession,
    block_schema_id: UUID,
) -> Union[BlockSchema, None]:
    """
    Reads a block schema by id. Will reconstruct the block schema's fields attribute
    to include block schema references.

    Args:
        session: A database session
        block_schema_id: a block_schema id

    Returns:
        orm_models..BlockSchema: the block_schema
    """

    # Construction of a recursive query which returns the specified block schema
    # along with and nested block schemas coupled with the ID of their parent schema
    # the key that they reside under.
    block_schema_references_query = (
        sa.select(db.BlockSchemaReference)
        .select_from(db.BlockSchemaReference)
        .filter_by(parent_block_schema_id=block_schema_id)
        .cte("block_schema_references", recursive=True)
    )
    block_schema_references_join = (
        sa.select(db.BlockSchemaReference)
        .select_from(db.BlockSchemaReference)
        .join(
            block_schema_references_query,
            db.BlockSchemaReference.parent_block_schema_id
            == block_schema_references_query.c.reference_block_schema_id,
        )
    )
    recursive_block_schema_references_cte = block_schema_references_query.union_all(
        block_schema_references_join
    )
    nested_block_schemas_query = (
        sa.select(
            db.BlockSchema,
            recursive_block_schema_references_cte.c.name,
            recursive_block_schema_references_cte.c.parent_block_schema_id,
        )
        .select_from(db.BlockSchema)
        .join(
            recursive_block_schema_references_cte,
            db.BlockSchema.id
            == recursive_block_schema_references_cte.c.reference_block_schema_id,
            isouter=True,
        )
        .filter(
            sa.or_(
                db.BlockSchema.id == block_schema_id,
                recursive_block_schema_references_cte.c.parent_block_schema_id.is_not(
                    None
                ),
            )
        )
    )
    result = await session.execute(nested_block_schemas_query)

    return _construct_full_block_schema(result.all())  # type: ignore[arg-type]


def _construct_full_block_schema(
    block_schemas_with_references: List[
        Tuple[BlockSchema, Optional[str], Optional[UUID]]
    ],
    root_block_schema: Optional[BlockSchema] = None,
) -> Optional[BlockSchema]:
    """
    Takes a list of block schemas along with reference information and reconstructs
    the root block schema's fields attribute to contain block schema references for
    client consumption.

    Args:
        block_schema_with_references: A list of tuples with the structure:
            - A block schema object
            - The name the block schema lives under in the parent block schema
            - The ID of the block schema's parent block schema
        root_block_schema: Optional block schema to start traversal. Will attempt to
            determine root block schema if not provided.

    Returns:
        BlockSchema: A block schema with a fully reconstructed fields attribute
    """
    if len(block_schemas_with_references) == 0:
        return None
    root_block_schema = (
        copy(root_block_schema)
        if root_block_schema is not None
        else _find_root_block_schema(block_schemas_with_references)
    )
    if root_block_schema is None:
        raise ValueError(
            "Unable to determine root block schema during schema reconstruction."
        )
    root_block_schema.fields = _construct_block_schema_fields_with_block_references(
        root_block_schema, block_schemas_with_references
    )
    definitions = _construct_block_schema_spec_definitions(
        root_block_schema, block_schemas_with_references
    )
    # Definitions for non block object may already exist in the block schema OpenAPI
    # spec, so we need to combine block and non-block definitions.
    if definitions or root_block_schema.fields.get("definitions"):
        root_block_schema.fields["definitions"] = {
            **root_block_schema.fields.get("definitions", {}),
            **definitions,
        }
    return root_block_schema


def _find_root_block_schema(
    block_schemas_with_references: List[
        Tuple[BlockSchema, Optional[str], Optional[UUID]]
    ],
) -> Union[BlockSchema, None]:
    """
    Attempts to find the root block schema from a list of block schemas
    with references. Returns None if a root block schema is not found.
    Returns only the first potential root block schema if multiple are found.
    """
    return next(
        (
            copy(block_schema)
            for (
                block_schema,
                _,
                parent_block_schema_id,
            ) in block_schemas_with_references
            if parent_block_schema_id is None
        ),
        None,
    )


def _construct_block_schema_spec_definitions(
    root_block_schema: BlockSchema,
    block_schemas_with_references: List[
        Tuple[BlockSchema, Optional[str], Optional[UUID]]
    ],
) -> dict[str, Any]:
    """
    Constructs field definitions for a block schema based on the nested block schemas
    as defined in the block_schemas_with_references list.
    """
    definitions: dict[str, Any] = {}
    for _, block_schema_references in root_block_schema.fields[
        "block_schema_references"
    ].items():
        block_schema_references = (
            block_schema_references
            if isinstance(block_schema_references, list)
            else [block_schema_references]
        )
        for block_schema_reference in block_schema_references:
            child_block_schema = _find_block_schema_via_checksum(
                block_schemas_with_references,
                block_schema_reference["block_schema_checksum"],
            )

            if child_block_schema is not None:
                child_block_schema = _construct_full_block_schema(
                    block_schemas_with_references=block_schemas_with_references,
                    root_block_schema=child_block_schema,
                )
                assert child_block_schema
                definitions = _add_block_schemas_fields_to_definitions(
                    definitions, child_block_schema
                )
    return definitions


def _find_block_schema_via_checksum(
    block_schemas_with_references: List[
        Tuple[BlockSchema, Optional[str], Optional[UUID]]
    ],
    checksum: str,
) -> Optional[BlockSchema]:
    """Attempt to find a block schema via a given checksum. Returns None if not found."""
    return next(
        (
            block_schema
            for block_schema, _, _ in block_schemas_with_references
            if block_schema.checksum == checksum
        ),
        None,
    )


def _add_block_schemas_fields_to_definitions(
    definitions: Dict, child_block_schema: BlockSchema
) -> dict[str, Any]:
    """
    Returns a new definitions dict with the fields of a block schema and it's child
    block schemas added to the existing definitions.
    """
    block_schema_title = child_block_schema.fields.get("title")
    if block_schema_title is not None:
        # Definitions are declared as a flat dict, so we pop off definitions
        # from child schemas and add them to the parent definitions dict
        child_definitions = child_block_schema.fields.pop("definitions", {})
        return {
            **definitions,
            **{block_schema_title: child_block_schema.fields},
            **child_definitions,
        }
    else:
        return definitions


def _construct_block_schema_fields_with_block_references(
    parent_block_schema: BlockSchema,
    block_schemas_with_references: List[
        Tuple[BlockSchema, Optional[str], Optional[UUID]]
    ],
) -> dict[str, Any]:
    """
    Constructs the block_schema_references in a block schema's fields attributes. Returns
    a copy of the block schema with block_schema_references added.

    Args:
        parent_block_schema: The block schema that needs block references populated.
        block_schema_with_references: A list of tuples with the structure:
            - A block schema object
            - The name the block schema lives under in the parent block schema
            - The ID of the block schema's parent block schema

    Returns:
        Dict: Block schema fields with block schema references added.

    """
    block_schema_fields_copy = {
        **parent_block_schema.fields,
        "block_schema_references": {},
    }
    for (
        nested_block_schema,
        name,
        parent_block_schema_id,
    ) in block_schemas_with_references:
        if parent_block_schema_id == parent_block_schema.id:
            assert nested_block_schema.block_type, (
                f"{nested_block_schema} has no block type"
            )

            new_block_schema_reference = {
                "block_schema_checksum": nested_block_schema.checksum,
                "block_type_slug": nested_block_schema.block_type.slug,
            }
            # A block reference for this key does not yet exist
            if name not in block_schema_fields_copy["block_schema_references"]:
                block_schema_fields_copy["block_schema_references"][name] = (
                    new_block_schema_reference
                )
            else:
                # List of block references for this key already exist and the block
                # reference that we are attempting add isn't present
                if (
                    isinstance(
                        block_schema_fields_copy["block_schema_references"][name],
                        list,
                    )
                    and new_block_schema_reference
                    not in block_schema_fields_copy["block_schema_references"][name]
                ):
                    block_schema_fields_copy["block_schema_references"][name].append(
                        new_block_schema_reference
                    )
                # A single block reference for this key already exists and it does not
                # match the block reference that we are attempting to add
                elif (
                    block_schema_fields_copy["block_schema_references"][name]
                    != new_block_schema_reference
                ):
                    block_schema_fields_copy["block_schema_references"][name] = [
                        block_schema_fields_copy["block_schema_references"][name],
                        new_block_schema_reference,
                    ]
    return block_schema_fields_copy


@db_injector
async def read_block_schemas(
    db: PrefectDBInterface,
    session: AsyncSession,
    block_schema_filter: Optional[schemas.filters.BlockSchemaFilter] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[BlockSchema]:
    """
    Reads block schemas, optionally filtered by type or name.

    Args:
        session: A database session
        block_schema_filter: a block schema filter object
        limit (int): query limit
        offset (int): query offset

    Returns:
        List[orm_models.BlockSchema]: the block_schemas
    """
    # schemas are ordered by `created DESC` to get the most recently created
    # ones first (and to facilitate getting the newest one with `limit=1`).
    filtered_block_schemas_query = select(db.BlockSchema.id).order_by(
        db.BlockSchema.created.desc()
    )

    if block_schema_filter:
        filtered_block_schemas_query = filtered_block_schemas_query.where(
            block_schema_filter.as_sql_filter()
        )

    if offset is not None:
        filtered_block_schemas_query = filtered_block_schemas_query.offset(offset)
    if limit is not None:
        filtered_block_schemas_query = filtered_block_schemas_query.limit(limit)

    filtered_block_schema_ids = (
        (await session.execute(filtered_block_schemas_query)).scalars().unique().all()
    )

    block_schema_references_query = (
        sa.select(db.BlockSchemaReference)
        .select_from(db.BlockSchemaReference)
        .filter(
            db.BlockSchemaReference.parent_block_schema_id.in_(
                filtered_block_schemas_query
            )
        )
        .cte("block_schema_references", recursive=True)
    )
    block_schema_references_join = (
        sa.select(db.BlockSchemaReference)
        .select_from(db.BlockSchemaReference)
        .join(
            block_schema_references_query,
            db.BlockSchemaReference.parent_block_schema_id
            == block_schema_references_query.c.reference_block_schema_id,
        )
    )
    recursive_block_schema_references_cte = block_schema_references_query.union_all(
        block_schema_references_join
    )

    nested_block_schemas_query = (
        sa.select(
            db.BlockSchema,
            recursive_block_schema_references_cte.c.name,
            recursive_block_schema_references_cte.c.parent_block_schema_id,
        )
        .select_from(db.BlockSchema)
        # in order to reconstruct nested block schemas efficiently, we need to visit them
        # in the order they were created (so that we guarantee that nested/referenced schemas)
        # have already been seen. Therefore this second query sorts by created ASC
        .order_by(db.BlockSchema.created.asc())
        .join(
            recursive_block_schema_references_cte,
            db.BlockSchema.id
            == recursive_block_schema_references_cte.c.reference_block_schema_id,
            isouter=True,
        )
        .filter(
            sa.or_(
                db.BlockSchema.id.in_(filtered_block_schemas_query),
                recursive_block_schema_references_cte.c.parent_block_schema_id.is_not(
                    None
                ),
            )
        )
    )

    block_schemas_with_references = (
        (await session.execute(nested_block_schemas_query)).unique().all()
    )
    fully_constructed_block_schemas = []
    visited_block_schema_ids = []
    for root_block_schema, _, _ in block_schemas_with_references:
        if (
            root_block_schema.id in filtered_block_schema_ids
            and root_block_schema.id not in visited_block_schema_ids
        ):
            constructed = _construct_full_block_schema(
                block_schemas_with_references=block_schemas_with_references,  # type: ignore[arg-type]
                root_block_schema=root_block_schema,
            )
            assert constructed
            fully_constructed_block_schemas.append(constructed)
            visited_block_schema_ids.append(root_block_schema.id)

    # because we reconstructed schemas ordered by created ASC, we
    # reverse the final output to restore created DESC
    return list(reversed(fully_constructed_block_schemas))


@db_injector
async def read_block_schema_by_checksum(
    db: PrefectDBInterface,
    session: AsyncSession,
    checksum: str,
    version: Optional[str] = None,
) -> Optional[BlockSchema]:
    """
    Reads a block_schema by checksum. Will reconstruct the block schema's fields
    attribute to include block schema references.

    Args:
        session: A database session
        checksum: a block_schema checksum
        version: A block_schema version

    Returns:
        orm_models.BlockSchema: the block_schema
    """
    # Construction of a recursive query which returns the specified block schema
    # along with and nested block schemas coupled with the ID of their parent schema
    # the key that they reside under.

    # The same checksum with different versions can occur in the DB. Return only the
    # most recently created one.
    root_block_schema_query = (
        sa.select(db.BlockSchema)
        .filter_by(checksum=checksum)
        .order_by(db.BlockSchema.created.desc())
        .limit(1)
    )

    if version is not None:
        root_block_schema_query = root_block_schema_query.filter_by(version=version)

    root_block_schema_cte = root_block_schema_query.cte("root_block_schema")

    block_schema_references_query = (
        sa.select(db.BlockSchemaReference)
        .select_from(db.BlockSchemaReference)
        .filter_by(parent_block_schema_id=root_block_schema_cte.c.id)
        .cte("block_schema_references", recursive=True)
    )
    block_schema_references_join = (
        sa.select(db.BlockSchemaReference)
        .select_from(db.BlockSchemaReference)
        .join(
            block_schema_references_query,
            db.BlockSchemaReference.parent_block_schema_id
            == block_schema_references_query.c.reference_block_schema_id,
        )
    )
    recursive_block_schema_references_cte = block_schema_references_query.union_all(
        block_schema_references_join
    )
    nested_block_schemas_query = (
        sa.select(
            db.BlockSchema,
            recursive_block_schema_references_cte.c.name,
            recursive_block_schema_references_cte.c.parent_block_schema_id,
        )
        .select_from(db.BlockSchema)
        .join(
            recursive_block_schema_references_cte,
            db.BlockSchema.id
            == recursive_block_schema_references_cte.c.reference_block_schema_id,
            isouter=True,
        )
        .filter(
            sa.or_(
                db.BlockSchema.id == root_block_schema_cte.c.id,
                recursive_block_schema_references_cte.c.parent_block_schema_id.is_not(
                    None
                ),
            )
        )
    )
    result = await session.execute(nested_block_schemas_query)
    return _construct_full_block_schema(result.all())  # type: ignore[arg-type]


@db_injector
async def read_available_block_capabilities(
    db: PrefectDBInterface,
    session: AsyncSession,
) -> List[str]:
    """
    Retrieves a list of all available block capabilities.

    Args:
        session: A database session.

    Returns:
        List[str]: List of all available block capabilities.
    """
    query = sa.select(
        db.queries.json_arr_agg(
            db.queries.cast_to_json(db.BlockSchema.capabilities.distinct())
        )
    )
    capability_combinations = (await session.execute(query)).scalars().first() or list()
    if db.queries.uses_json_strings and isinstance(capability_combinations, str):
        capability_combinations = json.loads(capability_combinations)
    return list({c for capabilities in capability_combinations for c in capabilities})


@db_injector
async def create_block_schema_reference(
    db: PrefectDBInterface,
    session: AsyncSession,
    block_schema_reference: schemas.core.BlockSchemaReference,
) -> Union[orm_models.BlockSchemaReference, None]:
    """
    Retrieves a list of all available block capabilities.

    Args:
        session: A database session.
        block_schema_reference: A block schema reference object.

    Returns:
        orm_models.BlockSchemaReference: The created BlockSchemaReference
    """
    query_stmt = sa.select(db.BlockSchemaReference).where(
        db.BlockSchemaReference.name == block_schema_reference.name,
        db.BlockSchemaReference.parent_block_schema_id
        == block_schema_reference.parent_block_schema_id,
        db.BlockSchemaReference.reference_block_schema_id
        == block_schema_reference.reference_block_schema_id,
    )

    existing_reference = (await session.execute(query_stmt)).scalar()
    if existing_reference:
        return existing_reference

    insert_stmt = db.queries.insert(db.BlockSchemaReference).values(
        **block_schema_reference.model_dump_for_orm(
            exclude_unset=True, exclude={"created", "updated"}
        )
    )
    await session.execute(insert_stmt)

    result = await session.execute(
        sa.select(db.BlockSchemaReference).where(
            db.BlockSchemaReference.id == block_schema_reference.id
        )
    )
    return result.scalar()



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, NoReturn

import rich

from prefect.logging import get_logger
from prefect.server.events.schemas.events import ReceivedEvent
from prefect.server.services.base import RunInAllServers, Service
from prefect.server.utilities.messaging import Consumer, Message, create_consumer
from prefect.settings.context import get_current_settings
from prefect.settings.models.server.services import ServicesBaseSetting
from prefect.types._datetime import now

if TYPE_CHECKING:
    import logging

logger: "logging.Logger" = get_logger(__name__)


class EventLogger(RunInAllServers, Service):
    """A debugging service that logs events to the console as they arrive."""

    consumer_task: asyncio.Task[None] | None = None

    @classmethod
    def service_settings(cls) -> ServicesBaseSetting:
        return get_current_settings().server.services.event_logger

    async def start(self) -> NoReturn:
        assert self.consumer_task is None, "Logger already started"
        self.consumer: Consumer = create_consumer("events")

        console = rich.console.Console()

        async def handler(message: Message):
            right_now = now("UTC")
            event: ReceivedEvent = ReceivedEvent.model_validate_json(message.data)

            console.print(
                "Event:",
                str(event.id).partition("-")[0],
                f"{event.occurred.isoformat()}",
                f" ({(event.occurred - right_now).total_seconds():>6,.2f})",
                f"\\[[bold green]{event.event}[/]]",
                event.resource.id,
            )
            console.file.flush()

        self.consumer_task = asyncio.create_task(self.consumer.run(handler))
        logger.debug("Event logger started")

        try:
            await self.consumer_task
        except asyncio.CancelledError:
            pass

    async def stop(self) -> None:
        assert self.consumer_task is not None, "Logger not started"
        self.consumer_task.cancel()
        try:
            await self.consumer_task
        except asyncio.CancelledError:
            pass
        finally:
            self.consumer_task = None
        logger.debug("Event logger stopped")



from __future__ import annotations

import asyncio
import copy
import threading
from collections import defaultdict
from collections.abc import AsyncGenerator, Iterable, Mapping, MutableMapping
from contextlib import asynccontextmanager
from dataclasses import asdict, dataclass
from datetime import timedelta
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union
from uuid import uuid4

import anyio
from cachetools import TTLCache
from exceptiongroup import BaseExceptionGroup  # novermin
from pydantic_core import to_json
from typing_extensions import Self

from prefect.logging import get_logger
from prefect.server.utilities.messaging import Cache as _Cache
from prefect.server.utilities.messaging import Consumer as _Consumer
from prefect.server.utilities.messaging import Message, MessageHandler, StopConsumer
from prefect.server.utilities.messaging import Publisher as _Publisher
from prefect.settings.context import get_current_settings

if TYPE_CHECKING:
    import logging

logger: "logging.Logger" = get_logger(__name__)

# Simple global counters by topic with thread-safe access
_metrics_lock: threading.Lock | None = None
METRICS: dict[str, dict[str, int]] = defaultdict(
    lambda: {
        "published": 0,
        "retried": 0,
        "consumed": 0,
    }
)


async def log_metrics_periodically(interval: float = 2.0) -> None:
    if _metrics_lock is None:
        return
    while True:
        await asyncio.sleep(interval)
        with _metrics_lock:
            for topic, data in METRICS.items():
                if data["published"] == 0:
                    continue
                depth = data["published"] - data["consumed"]
                logger.debug(
                    "Topic=%r | published=%d consumed=%d retried=%d depth=%d",
                    topic,
                    data["published"],
                    data["consumed"],
                    data["retried"],
                    depth,
                )


async def update_metric(topic: str, key: str, amount: int = 1) -> None:
    global _metrics_lock
    if _metrics_lock is None:
        _metrics_lock = threading.Lock()
    with _metrics_lock:
        METRICS[topic][key] += amount


@dataclass
class MemoryMessage:
    data: Union[bytes, str]
    attributes: Mapping[str, Any]
    retry_count: int = 0


class Subscription:
    """
    A subscription to a topic.

    Messages are delivered to the subscription's queue and retried up to a
    maximum number of times. If a message cannot be delivered after the maximum
    number of retries it is moved to the dead letter queue.

    The dead letter queue is a directory of JSON files containing the serialized
    message.

    Messages remain in the dead letter queue until they are removed manually.

    Attributes:
        topic: The topic that the subscription receives messages from.
        max_retries: The maximum number of times a message will be retried for
            this subscription.
        dead_letter_queue_path: The path to the dead letter queue folder.
    """

    def __init__(
        self,
        topic: "Topic",
        max_retries: int = 3,
        dead_letter_queue_path: Path | str | None = None,
    ) -> None:
        self.topic = topic
        self.max_retries = max_retries
        self.dead_letter_queue_path: Path = (
            Path(dead_letter_queue_path)
            if dead_letter_queue_path
            else get_current_settings().home / "dlq"
        )
        self._queue: asyncio.Queue[MemoryMessage] = asyncio.Queue()
        self._retry: asyncio.Queue[MemoryMessage] = asyncio.Queue()

    async def deliver(self, message: MemoryMessage) -> None:
        """
        Deliver a message to the subscription's queue.

        Args:
            message: The message to deliver.
        """
        await self._queue.put(message)
        await update_metric(self.topic.name, "published")
        logger.debug(
            "Delivered message to topic=%r queue_size=%d retry_queue_size=%d",
            self.topic.name,
            self._queue.qsize(),
            self._retry.qsize(),
        )

    async def retry(self, message: MemoryMessage) -> None:
        """
        Place a message back on the retry queue.

        If the message has retried more than the maximum number of times it is
        moved to the dead letter queue.

        Args:
            message: The message to retry.
        """
        message.retry_count += 1
        if message.retry_count > self.max_retries:
            logger.warning(
                "Message failed after %d retries and will be moved to the dead letter queue",
                message.retry_count,
                extra={"event_message": message},
            )
            await self.send_to_dead_letter_queue(message)
        else:
            await self._retry.put(message)
            await update_metric(self.topic.name, "retried")
            logger.debug(
                "Retried message on topic=%r retry_count=%d queue_size=%d retry_queue_size=%d",
                self.topic.name,
                message.retry_count,
                self._queue.qsize(),
                self._retry.qsize(),
            )

    async def get(self) -> MemoryMessage:
        """
        Get a message from the subscription's queue.
        """
        if not self._retry.empty():
            return await self._retry.get()
        return await self._queue.get()

    async def send_to_dead_letter_queue(self, message: MemoryMessage) -> None:
        """
        Send a message to the dead letter queue.

        The dead letter queue is a directory of JSON files containing the
        serialized messages.

        Args:
            message: The message to send to the dead letter queue.
        """
        self.dead_letter_queue_path.mkdir(parents=True, exist_ok=True)
        try:
            await anyio.Path(self.dead_letter_queue_path / uuid4().hex).write_bytes(
                to_json(asdict(message))
            )
        except Exception as e:
            logger.warning("Failed to write message to dead letter queue", exc_info=e)


class Topic:
    _topics: dict[str, "Topic"] = {}

    name: str
    _subscriptions: list[Subscription]

    def __init__(self, name: str) -> None:
        self.name = name
        self._subscriptions = []

    @classmethod
    def by_name(cls, name: str) -> "Topic":
        try:
            return cls._topics[name]
        except KeyError:
            topic = cls(name)
            cls._topics[name] = topic
            return topic

    @classmethod
    def clear_all(cls) -> None:
        for topic in cls._topics.values():
            topic.clear()
        cls._topics = {}

    def subscribe(self, **subscription_kwargs: Any) -> Subscription:
        subscription = Subscription(self, **subscription_kwargs)
        self._subscriptions.append(subscription)
        return subscription

    def unsubscribe(self, subscription: Subscription) -> None:
        self._subscriptions.remove(subscription)

    def clear(self) -> None:
        for subscription in self._subscriptions:
            self.unsubscribe(subscription)
        self._subscriptions = []

    async def publish(self, message: MemoryMessage) -> None:
        for subscription in self._subscriptions:
            # Ensure that each subscription gets its own copy of the message
            await subscription.deliver(copy.deepcopy(message))


@asynccontextmanager
async def break_topic():
    from unittest import mock

    publishing_mock = mock.AsyncMock(side_effect=ValueError("oops"))

    with mock.patch(
        "prefect.server.utilities.messaging.memory.Topic.publish",
        publishing_mock,
    ):
        yield


M = TypeVar("M", bound=Message)


class Cache(_Cache):
    _recently_seen_messages: MutableMapping[str, bool] = TTLCache(
        maxsize=1000,
        ttl=timedelta(minutes=5).total_seconds(),
    )

    async def clear_recently_seen_messages(self) -> None:
        self._recently_seen_messages.clear()

    async def without_duplicates(
        self, attribute: str, messages: Iterable[M]
    ) -> list[M]:
        messages_with_attribute: list[M] = []
        messages_without_attribute: list[M] = []

        for m in messages:
            if not m.attributes or attribute not in m.attributes:
                logger.warning(
                    "Message is missing deduplication attribute %r",
                    attribute,
                    extra={"event_message": m},
                )
                messages_without_attribute.append(m)
                continue

            if self._recently_seen_messages.get(m.attributes[attribute]):
                continue

            self._recently_seen_messages[m.attributes[attribute]] = True
            messages_with_attribute.append(m)

        return messages_with_attribute + messages_without_attribute

    async def forget_duplicates(self, attribute: str, messages: Iterable[M]) -> None:
        for m in messages:
            if not m.attributes or attribute not in m.attributes:
                logger.warning(
                    "Message is missing deduplication attribute %r",
                    attribute,
                    extra={"event_message": m},
                )
                continue
            self._recently_seen_messages.pop(m.attributes[attribute], None)


class Publisher(_Publisher):
    def __init__(self, topic: str, cache: Cache, deduplicate_by: Optional[str] = None):
        self.topic: Topic = Topic.by_name(topic)
        self.deduplicate_by = deduplicate_by
        self._cache = cache

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        return None

    async def publish_data(self, data: bytes, attributes: Mapping[str, str]) -> None:
        to_publish = [MemoryMessage(data, attributes)]
        if self.deduplicate_by:
            to_publish = await self._cache.without_duplicates(
                self.deduplicate_by, to_publish
            )

        try:
            for message in to_publish:
                await self.topic.publish(message)
        except Exception:
            if self.deduplicate_by:
                await self._cache.forget_duplicates(self.deduplicate_by, to_publish)
            raise


class Consumer(_Consumer):
    def __init__(
        self,
        topic: str,
        subscription: Optional[Subscription] = None,
        concurrency: int = 2,
    ):
        self.topic: Topic = Topic.by_name(topic)
        if not subscription:
            subscription = self.topic.subscribe()
        assert subscription.topic is self.topic
        self.subscription = subscription
        self.concurrency = concurrency

    async def run(self, handler: MessageHandler) -> None:
        try:
            async with anyio.create_task_group() as tg:
                for _ in range(self.concurrency):
                    tg.start_soon(self._consume_loop, handler)
        except BaseExceptionGroup as group:  # novermin
            if all(isinstance(exc, StopConsumer) for exc in group.exceptions):
                logger.debug("StopConsumer received")
                return  # Exit cleanly when all tasks stop
            # Re-raise if any non-StopConsumer exceptions
            raise group

    async def _consume_loop(self, handler: MessageHandler) -> None:
        while True:
            message = await self.subscription.get()
            try:
                await handler(message)
                await update_metric(self.topic.name, "consumed")
            except StopConsumer as e:
                if not e.ack:
                    await self.subscription.retry(message)
                raise  # Propagate to task group
            except Exception:
                logger.exception("Failed in consume_loop")
                await self.subscription.retry(message)


@asynccontextmanager
async def ephemeral_subscription(topic: str) -> AsyncGenerator[Mapping[str, Any], None]:
    subscription = Topic.by_name(topic).subscribe()
    try:
        yield {"topic": topic, "subscription": subscription}
    finally:
        Topic.by_name(topic).unsubscribe(subscription)



import importlib
import sys
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .actions import BlockTypeUpdate, StateCreate
    from .objects import (
        DEFAULT_BLOCK_SCHEMA_VERSION,
        BlockDocument,
        BlockSchema,
        BlockType,
        FlowRun,
        FlowRunPolicy,
        State,
        StateDetails,
        StateType,
        TaskRun,
        TaskRunInput,
        TaskRunPolicy,
        TaskRunResult,
        Workspace,
    )
    from .responses import (
        OrchestrationResult,
        SetStateStatus,
        StateAbortDetails,
        StateAcceptDetails,
        StateRejectDetails,
    )

_public_api = {
    "BlockDocument": (__package__, ".objects"),
    "BlockSchema": (__package__, ".objects"),
    "BlockType": (__package__, ".objects"),
    "BlockTypeUpdate": (__package__, ".actions"),
    "DEFAULT_BLOCK_SCHEMA_VERSION": (__package__, ".objects"),
    "FlowRun": (__package__, ".objects"),
    "FlowRunPolicy": (__package__, ".objects"),
    "OrchestrationResult": (__package__, ".responses"),
    "SetStateStatus": (__package__, ".responses"),
    "State": (__package__, ".objects"),
    "StateAbortDetails": (__package__, ".responses"),
    "StateAcceptDetails": (__package__, ".responses"),
    "StateCreate": (__package__, ".actions"),
    "StateDetails": (__package__, ".objects"),
    "StateRejectDetails": (__package__, ".responses"),
    "StateType": (__package__, ".objects"),
    "TaskRun": (__package__, ".objects"),
    "TaskRunInput": (__package__, ".objects"),
    "TaskRunPolicy": (__package__, ".objects"),
    "TaskRunResult": (__package__, ".objects"),
    "Workspace": (__package__, ".objects"),
}

__all__ = [
    "BlockDocument",
    "BlockSchema",
    "BlockType",
    "BlockTypeUpdate",
    "DEFAULT_BLOCK_SCHEMA_VERSION",
    "FlowRun",
    "FlowRunPolicy",
    "OrchestrationResult",
    "SetStateStatus",
    "State",
    "StateAbortDetails",
    "StateAcceptDetails",
    "StateCreate",
    "StateDetails",
    "StateRejectDetails",
    "StateType",
    "TaskRun",
    "TaskRunInput",
    "TaskRunPolicy",
    "TaskRunResult",
    "Workspace",
]


def __getattr__(attr_name: str) -> Any:
    try:
        if (dynamic_attr := _public_api.get(attr_name)) is None:
            raise AttributeError(f"module {__name__} has no attribute {attr_name}")

        package, mname = dynamic_attr
        module = importlib.import_module(mname, package=package)
        return getattr(module, attr_name)
    except ModuleNotFoundError as ex:
        mname, _, attr = (ex.name or "").rpartition(".")
        ctx = {"name": mname, "obj": attr} if sys.version_info >= (3, 10) else {}
        raise AttributeError(f"module {mname} has no attribute {attr}", **ctx) from ex



from typing import Optional
from uuid import UUID

from pydantic import BaseModel, Field


class CreatedBy(BaseModel):
    id: Optional[UUID] = Field(
        default=None, description="The id of the creator of the object."
    )
    type: Optional[str] = Field(
        default=None, description="The type of the creator of the object."
    )
    display_value: Optional[str] = Field(
        default=None, description="The display value for the creator."
    )


class UpdatedBy(BaseModel):
    id: Optional[UUID] = Field(
        default=None, description="The id of the updater of the object."
    )
    type: Optional[str] = Field(
        default=None, description="The type of the updater of the object."
    )
    display_value: Optional[str] = Field(
        default=None, description="The display value for the updater."
    )



"""Adds heartbeat_interval_seconds column to worker table

Revision ID: c2d001b7dd06
Revises: 5b0bd3b41a23
Create Date: 2023-09-06 08:47:29.381718

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c2d001b7dd06"
down_revision = "5b0bd3b41a23"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("PRAGMA foreign_keys=OFF")

    with op.batch_alter_table("worker", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("heartbeat_interval_seconds", sa.Integer(), nullable=True)
        )

    op.execute("PRAGMA foreign_keys=ON")


def downgrade():
    op.execute("PRAGMA foreign_keys=OFF")

    with op.batch_alter_table("worker", schema=None) as batch_op:
        batch_op.drop_column("heartbeat_interval_seconds")

    op.execute("PRAGMA foreign_keys=ON")



from __future__ import annotations

from typing import TYPE_CHECKING

import httpx

from prefect.client.orchestration.base import BaseAsyncClient, BaseClient
from prefect.exceptions import ObjectNotFound

if TYPE_CHECKING:
    from prefect.client.schemas.actions import (
        VariableCreate,
        VariableUpdate,
    )
    from prefect.client.schemas.objects import (
        Variable,
    )


class VariableClient(BaseClient):
    def create_variable(self, variable: "VariableCreate") -> "Variable":
        """
        Creates an variable with the provided configuration.

        Args:
            variable: Desired configuration for the new variable.
        Returns:
            Information about the newly created variable.
        """
        response = self._client.post(
            "/variables/",
            json=variable.model_dump(mode="json", exclude_unset=True),
        )
        from prefect.client.schemas.objects import Variable

        return Variable.model_validate(response.json())

    def read_variable_by_name(self, name: str) -> "Variable | None":
        """Reads a variable by name. Returns None if no variable is found."""
        try:
            response = self.request(
                "GET", "/variables/name/{name}", path_params={"name": name}
            )
            from prefect.client.schemas.objects import Variable

            return Variable(**response.json())
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                return None
            else:
                raise

    def read_variables(self, limit: int | None = None) -> list["Variable"]:
        """Reads all variables."""
        response = self.request("POST", "/variables/filter", json={"limit": limit})
        from prefect.client.schemas.objects import Variable

        return Variable.model_validate_list(response.json())

    def update_variable(self, variable: "VariableUpdate") -> None:
        """
        Updates a variable with the provided configuration.

        Args:
            variable: Desired configuration for the updated variable.
        Returns:
            Information about the updated variable.
        """
        self._client.patch(
            f"/variables/name/{variable.name}",
            json=variable.model_dump(mode="json", exclude_unset=True),
        )
        return None

    def delete_variable_by_name(self, name: str) -> None:
        """Deletes a variable by name."""
        try:
            self.request(
                "DELETE",
                "/variables/name/{name}",
                path_params={"name": name},
            )
            return None
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                raise ObjectNotFound(http_exc=e) from e
            else:
                raise


class VariableAsyncClient(BaseAsyncClient):
    async def create_variable(self, variable: "VariableCreate") -> "Variable":
        """Creates a variable with the provided configuration."""
        response = await self._client.post(
            "/variables/",
            json=variable.model_dump(mode="json", exclude_unset=True),
        )
        from prefect.client.schemas.objects import Variable

        return Variable.model_validate(response.json())

    async def read_variable_by_name(self, name: str) -> "Variable | None":
        """Reads a variable by name. Returns None if no variable is found."""
        try:
            response = await self.request(
                "GET",
                "/variables/name/{name}",
                path_params={"name": name},
            )
            from prefect.client.schemas.objects import Variable

            return Variable.model_validate(response.json())
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                return None
            else:
                raise

    async def read_variables(self, limit: int | None = None) -> list["Variable"]:
        """Reads all variables."""
        response = await self.request(
            "POST", "/variables/filter", json={"limit": limit}
        )
        from prefect.client.schemas.objects import Variable

        return Variable.model_validate_list(response.json())

    async def update_variable(self, variable: "VariableUpdate") -> None:
        """
        Updates a variable with the provided configuration.

        Args:
            variable: Desired configuration for the updated variable.
        Returns:
            Information about the updated variable.
        """
        await self.request(
            "PATCH",
            "/variables/name/{name}",
            path_params={"name": variable.name},
            json=variable.model_dump(mode="json", exclude_unset=True),
        )
        return None

    async def delete_variable_by_name(self, name: str) -> None:
        """Deletes a variable by name."""
        try:
            await self.request(
                "DELETE",
                "/variables/name/{name}",
                path_params={"name": name},
            )
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                raise ObjectNotFound(http_exc=e) from e
            else:
                raise



import asyncio
from enum import Enum

import orjson
import typer
import websockets
from anyio import open_file

from prefect.cli._types import PrefectTyper
from prefect.cli._utilities import exit_with_error
from prefect.cli.root import app
from prefect.events import Event
from prefect.events.clients import (
    PrefectCloudAccountEventSubscriber,
    get_events_subscriber,
)

events_app: PrefectTyper = PrefectTyper(name="events", help="Stream events.")
app.add_typer(events_app, aliases=["event"])


class StreamFormat(str, Enum):
    json = "json"
    text = "text"


@events_app.command()
async def stream(
    format: StreamFormat = typer.Option(
        StreamFormat.json, "--format", help="Output format (json or text)"
    ),
    output_file: str = typer.Option(
        None, "--output-file", help="File to write events to"
    ),
    account: bool = typer.Option(
        False,
        "--account",
        help="Stream events for entire account, including audit logs",
    ),
    run_once: bool = typer.Option(False, "--run-once", help="Stream only one event"),
):
    """Subscribes to the event stream of a workspace, printing each event
    as it is received. By default, events are printed as JSON, but can be
    printed as text by passing `--format text`.
    """

    try:
        if account:
            events_subscriber = PrefectCloudAccountEventSubscriber()
        else:
            events_subscriber = get_events_subscriber()

        app.console.print("Subscribing to event stream...")
        async with events_subscriber as subscriber:
            async for event in subscriber:
                await handle_event(event, format, output_file)
                if run_once:
                    typer.Exit(0)
    except Exception as exc:
        handle_error(exc)


async def handle_event(event: Event, format: StreamFormat, output_file: str) -> None:
    if format == StreamFormat.json:
        event_data = orjson.dumps(event.model_dump(), default=str).decode()
    elif format == StreamFormat.text:
        event_data = f"{event.occurred.isoformat()} {event.event} {event.resource.id}"
    else:
        raise ValueError(f"Unknown format: {format}")
    if output_file:
        async with open_file(output_file, "a") as f:  # type: ignore
            await f.write(event_data + "\n")
    else:
        print(event_data)


def handle_error(exc: Exception) -> None:
    if isinstance(exc, websockets.exceptions.ConnectionClosedError):
        exit_with_error(f"Connection closed, retrying... ({exc})")
    elif isinstance(exc, (KeyboardInterrupt, asyncio.exceptions.CancelledError)):
        exit_with_error("Exiting...")
    elif isinstance(exc, (PermissionError)):
        exit_with_error(f"Error writing to file: {exc}")
    else:
        exit_with_error(f"An unexpected error occurred: {exc}")



from typing import TYPE_CHECKING
from prefect._internal.compatibility.migration import getattr_migration


if TYPE_CHECKING:
    from .flow_runs import run_deployment
    from .base import initialize_project
    from .runner import deploy

_public_api: dict[str, tuple[str, str]] = {
    "initialize_project": (__spec__.parent, ".base"),
    "run_deployment": (__spec__.parent, ".flow_runs"),
    "deploy": (__spec__.parent, ".runner"),
}

# Declare API for type-checkers
__all__ = ["initialize_project", "deploy", "run_deployment"]


def __getattr__(attr_name: str) -> object:
    dynamic_attr = _public_api.get(attr_name)
    if dynamic_attr is None:
        return getattr_migration(__name__)(attr_name)

    package, module_name = dynamic_attr

    from importlib import import_module

    if module_name == "__module__":
        return import_module(f".{attr_name}", package=package)
    else:
        module = import_module(module_name, package=package)
        return getattr(module, attr_name)



from typing import TYPE_CHECKING, Annotated, Optional

from httpx import HTTPStatusError
from pydantic import Field
from typing_extensions import TypedDict, Unpack

from prefect.client.orchestration.base import BaseAsyncClient, BaseClient
from prefect.exceptions import ObjectNotFound

if TYPE_CHECKING:
    from uuid import UUID

    from prefect.client.schemas.actions import ArtifactCreate, ArtifactUpdate
    from prefect.client.schemas.filters import (
        ArtifactCollectionFilter,
        ArtifactFilter,
        FlowRunFilter,
        TaskRunFilter,
    )
    from prefect.client.schemas.objects import Artifact, ArtifactCollection
    from prefect.client.schemas.sorting import ArtifactCollectionSort, ArtifactSort


class BaseArtifactReadParams(TypedDict, total=False):
    flow_run_filter: Annotated[Optional["FlowRunFilter"], Field(default=None)]
    task_run_filter: Annotated[Optional["TaskRunFilter"], Field(default=None)]
    limit: Annotated[Optional[int], Field(default=None)]
    offset: Annotated[int, Field(default=0)]


class ArtifactReadParams(BaseArtifactReadParams, total=False):
    artifact_filter: Annotated[Optional["ArtifactFilter"], Field(default=None)]
    sort: Annotated[Optional["ArtifactSort"], Field(default=None)]


class ArtifactCollectionReadParams(BaseArtifactReadParams, total=False):
    artifact_filter: Annotated[
        Optional["ArtifactCollectionFilter"], Field(default=None)
    ]
    sort: Annotated[Optional["ArtifactCollectionSort"], Field(default=None)]


class ArtifactClient(BaseClient):
    def create_artifact(self, artifact: "ArtifactCreate") -> "Artifact":
        response = self.request(
            "POST",
            "/artifacts/",
            json=artifact.model_dump(mode="json", exclude_unset=True),
        )
        from prefect.client.schemas.objects import Artifact

        return Artifact.model_validate(response.json())

    def update_artifact(self, artifact_id: "UUID", artifact: "ArtifactUpdate") -> None:
        self.request(
            "PATCH",
            "/artifacts/{id}",
            json=artifact.model_dump(mode="json", exclude_unset=True),
            path_params={"id": artifact_id},
        )
        return None

    def delete_artifact(self, artifact_id: "UUID") -> None:
        try:
            self.request(
                "DELETE",
                "/artifacts/{id}",
                path_params={"id": artifact_id},
            )
        except HTTPStatusError as e:
            if e.response.status_code == 404:
                raise ObjectNotFound(http_exc=e) from e
            else:
                raise
        return None

    def read_artifacts(
        self, **kwargs: Unpack["ArtifactReadParams"]
    ) -> list["Artifact"]:
        response = self.request(
            "POST",
            "/artifacts/filter",
            json={
                "artifacts": (
                    artifact_filter.model_dump(mode="json", exclude_unset=True)
                    if (artifact_filter := kwargs.get("artifact_filter"))
                    else None
                ),
                "flow_runs": (
                    flow_run_filter.model_dump(mode="json", exclude_unset=True)
                    if (flow_run_filter := kwargs.get("flow_run_filter"))
                    else None
                ),
                "task_runs": (
                    task_run_filter.model_dump(mode="json", exclude_unset=True)
                    if (task_run_filter := kwargs.get("task_run_filter"))
                    else None
                ),
                "limit": kwargs.get("limit"),
                "offset": kwargs.get("offset"),
                "sort": kwargs.get("sort"),
            },
        )
        from prefect.client.schemas.objects import Artifact

        return Artifact.model_validate_list(response.json())


class ArtifactAsyncClient(BaseAsyncClient):
    async def create_artifact(self, artifact: "ArtifactCreate") -> "Artifact":
        response = await self.request(
            "POST",
            "/artifacts/",
            json=artifact.model_dump(mode="json", exclude_unset=True),
        )
        from prefect.client.schemas.objects import Artifact

        return Artifact.model_validate(response.json())

    async def update_artifact(
        self, artifact_id: "UUID", artifact: "ArtifactUpdate"
    ) -> None:
        await self.request(
            "PATCH",
            "/artifacts/{id}",
            path_params={"id": artifact_id},
            json=artifact.model_dump(mode="json", exclude_unset=True),
        )
        return None

    async def read_artifacts(
        self, **kwargs: Unpack["ArtifactReadParams"]
    ) -> list["Artifact"]:
        response = await self.request(
            "POST",
            "/artifacts/filter",
            json={
                "artifacts": (
                    artifact_filter.model_dump(mode="json", exclude_unset=True)
                    if (artifact_filter := kwargs.get("artifact_filter"))
                    else None
                ),
                "flow_runs": (
                    flow_run_filter.model_dump(mode="json", exclude_unset=True)
                    if (flow_run_filter := kwargs.get("flow_run_filter"))
                    else None
                ),
                "task_runs": (
                    task_run_filter.model_dump(mode="json", exclude_unset=True)
                    if (task_run_filter := kwargs.get("task_run_filter"))
                    else None
                ),
                "limit": kwargs.get("limit", None),
                "offset": kwargs.get("offset", 0),
                "sort": kwargs.get("sort", None),
            },
        )
        from prefect.client.schemas.objects import Artifact

        return Artifact.model_validate_list(response.json())

    async def delete_artifact(self, artifact_id: "UUID") -> None:
        try:
            await self.request(
                "DELETE",
                "/artifacts/{id}",
                path_params={"id": artifact_id},
            )
        except HTTPStatusError as e:
            if e.response.status_code == 404:
                raise ObjectNotFound(http_exc=e) from e
            else:
                raise


class ArtifactCollectionClient(BaseClient):
    def read_latest_artifacts(
        self, **kwargs: Unpack["ArtifactCollectionReadParams"]
    ) -> list["ArtifactCollection"]:
        response = self.request(
            "POST",
            "/artifacts/latest/filter",
            json={
                "artifacts": (
                    artifact_filter.model_dump(mode="json", exclude_unset=True)
                    if (artifact_filter := kwargs.get("artifact_filter"))
                    else None
                ),
                "flow_runs": (
                    flow_run_filter.model_dump(mode="json", exclude_unset=True)
                    if (flow_run_filter := kwargs.get("flow_run_filter"))
                    else None
                ),
                "task_runs": (
                    task_run_filter.model_dump(mode="json", exclude_unset=True)
                    if (task_run_filter := kwargs.get("task_run_filter"))
                    else None
                ),
                "limit": kwargs.get("limit", None),
                "offset": kwargs.get("offset", 0),
                "sort": kwargs.get("sort", None),
            },
        )
        from prefect.client.schemas.objects import ArtifactCollection

        return ArtifactCollection.model_validate_list(response.json())


class ArtifactCollectionAsyncClient(BaseAsyncClient):
    async def read_latest_artifacts(
        self, **kwargs: Unpack["ArtifactCollectionReadParams"]
    ) -> list["ArtifactCollection"]:
        response = await self.request(
            "POST",
            "/artifacts/latest/filter",
            json={
                "artifacts": (
                    artifact_filter.model_dump(mode="json", exclude_unset=True)
                    if (artifact_filter := kwargs.get("artifact_filter"))
                    else None
                ),
                "flow_runs": (
                    flow_run_filter.model_dump(mode="json", exclude_unset=True)
                    if (flow_run_filter := kwargs.get("flow_run_filter"))
                    else None
                ),
                "task_runs": (
                    task_run_filter.model_dump(mode="json", exclude_unset=True)
                    if (task_run_filter := kwargs.get("task_run_filter"))
                    else None
                ),
                "limit": kwargs.get("limit", None),
                "offset": kwargs.get("offset", 0),
                "sort": kwargs.get("sort", None),
            },
        )
        from prefect.client.schemas.objects import ArtifactCollection

        return ArtifactCollection.model_validate_list(response.json())



import abc
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from dataclasses import dataclass
import importlib
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    runtime_checkable,
)
from collections.abc import AsyncGenerator, Awaitable, Iterable, Mapping
from typing_extensions import Self

from prefect.settings import PREFECT_MESSAGING_CACHE, PREFECT_MESSAGING_BROKER
from prefect.logging import get_logger

if TYPE_CHECKING:
    import logging

logger: "logging.Logger" = get_logger(__name__)


M = TypeVar("M", bound="Message", covariant=True)


class Message(Protocol):
    """
    A protocol representing a message sent to a message broker.
    """

    @property
    def data(self) -> Union[str, bytes]: ...

    @property
    def attributes(self) -> Mapping[str, Any]: ...


class Cache(abc.ABC):
    @abc.abstractmethod
    async def clear_recently_seen_messages(self) -> None: ...

    @abc.abstractmethod
    async def without_duplicates(
        self, attribute: str, messages: Iterable[M]
    ) -> list[M]: ...

    @abc.abstractmethod
    async def forget_duplicates(
        self, attribute: str, messages: Iterable[Message]
    ) -> None: ...


class Publisher(AbstractAsyncContextManager["Publisher"], abc.ABC):
    def __init__(
        self,
        topic: str,
        cache: Optional[Cache] = None,
        deduplicate_by: Optional[str] = None,
    ) -> None: ...

    @abc.abstractmethod
    async def publish_data(
        self, data: bytes, attributes: Mapping[str, str]
    ) -> None: ...

    @abc.abstractmethod
    async def __aenter__(self) -> Self: ...

    @abc.abstractmethod
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None: ...


@dataclass
class CapturedMessage:
    data: bytes
    attributes: Mapping[str, str]


class CapturingPublisher(Publisher):
    messages: list[CapturedMessage] = []
    deduplicate_by: Optional[str]

    def __init__(
        self,
        topic: str,
        cache: Optional[Cache] = None,
        deduplicate_by: Optional[str] = None,
    ) -> None:
        self.topic = topic
        self.cache: Cache = cache or create_cache()
        self.deduplicate_by = deduplicate_by

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    async def publish_data(self, data: bytes, attributes: Mapping[str, str]) -> None:
        to_publish = [CapturedMessage(data, attributes)]

        if self.deduplicate_by:
            to_publish = await self.cache.without_duplicates(
                self.deduplicate_by, to_publish
            )

        self.messages.extend(to_publish)


MessageHandler = Callable[[Message], Awaitable[None]]


class StopConsumer(Exception):
    """
    Exception to raise to stop a consumer.
    """

    def __init__(self, ack: bool = False):
        self.ack = ack


class Consumer(abc.ABC):
    """
    Abstract base class for consumers that receive messages from a message broker and
    call a handler function for each message received.
    """

    def __init__(self, topic: str, **kwargs: Any) -> None:
        self.topic = topic

    @abc.abstractmethod
    async def run(self, handler: MessageHandler) -> None:
        """Runs the consumer (indefinitely)"""
        ...


@runtime_checkable
class CacheModule(Protocol):
    Cache: type[Cache]


def create_cache() -> Cache:
    """
    Creates a new cache with the applications default settings.

    Returns:
        a new Cache instance
    """
    module = importlib.import_module(PREFECT_MESSAGING_CACHE.value())
    assert isinstance(module, CacheModule)

    return module.Cache()


@runtime_checkable
class BrokerModule(Protocol):
    Publisher: type[Publisher]
    Consumer: type[Consumer]
    ephemeral_subscription: Callable[
        [str], AbstractAsyncContextManager[Mapping[str, Any]]
    ]

    # Used for testing: a context manager that breaks the topic in a way that raises
    # a ValueError("oops") when attempting to publish a message.
    break_topic: Callable[[], AbstractAsyncContextManager[None]]


def create_publisher(
    topic: str, cache: Optional[Cache] = None, deduplicate_by: Optional[str] = None
) -> Publisher:
    """
    Creates a new publisher with the applications default settings.
    Args:
        topic: the topic to publish to
    Returns:
        a new Consumer instance
    """
    cache = cache or create_cache()

    module = importlib.import_module(PREFECT_MESSAGING_BROKER.value())
    assert isinstance(module, BrokerModule)
    return module.Publisher(topic, cache, deduplicate_by=deduplicate_by)


@asynccontextmanager
async def ephemeral_subscription(topic: str) -> AsyncGenerator[Mapping[str, Any], Any]:
    """
    Creates an ephemeral subscription to the given source, removing it when the context
    exits.
    """
    module = importlib.import_module(PREFECT_MESSAGING_BROKER.value())
    assert isinstance(module, BrokerModule)
    async with module.ephemeral_subscription(topic) as consumer_create_kwargs:
        yield consumer_create_kwargs


def create_consumer(topic: str, **kwargs: Any) -> Consumer:
    """
    Creates a new consumer with the applications default settings.
    Args:
        topic: the topic to consume from
    Returns:
        a new Consumer instance
    """
    module = importlib.import_module(PREFECT_MESSAGING_BROKER.value())
    assert isinstance(module, BrokerModule)
    return module.Consumer(topic, **kwargs)
